# #!/usr/bin/env python3
import sys, os, glob, time
import numpy as np
import random
import scipy
from subprocess import Popen
from time import sleep
from scipy.stats import spearmanr
from time import time
import fileinput,pickle
from scipy.stats import gaussian_kde
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import seaborn as sns
import h5py
from copy import deepcopy
try:
    from pyminer.common_functions import *
    from pyminer.pyminer_common_stats_functions import *
    from pyminer.pyminer_common_stats_functions import no_p_spear, get_Z_cutoff
except:
    from common_functions import *
    from pyminer_common_stats_functions import *
    from pyminer_common_stats_functions import no_p_spear, get_Z_cutoff
print(no_p_spear)

##################################################################################

def cutoff_check(r,p):
    global FPR, rho_cutoff, rho_cutoff_bool
    if rho_cutoff_bool:
        return(abs(r)>=rho_cutoff)
    else:
        return(p<=FPR)
    

#########

def get_spearman_h5(var_index):
    
    ## the offset is for a ! denoted header
    ## in general, this should be 1 because of the title line
    global temp, FPR, many_variables, IDlist, verbose, num_vars, infile
    
    output_lines = []
    cur_var_name1 = IDlist[var_index]
    for i in range(var_index+1,num_vars):
        if i%1000==0:
            print('\t',i)
        cur_var_name2 = IDlist[i]
        r,p=spearmanr(infile[var_index], infile[i])
        if cutoff_check(r,p):
            #print('found a relationship',r,p)
            output_lines.append([cur_var_name1, cur_var_name2, r])
            output_lines.append([cur_var_name2, cur_var_name1, r])
            #print(output_lines[-1])

    ## write status to temp log for this processes
    make_file("working on row "+str(var_index+1),log_file)

    
    if output_lines != []:
        return(output_lines)
    else:
        return(None)





##################################################################################
##############################################################
import argparse
parser = argparse.ArgumentParser()

parser.add_argument("-infile_path", '-i',
    help="the input dataset for making the adjacency list")
parser.add_argument("-adj_list_out", '-o',
    help="the ouptut adj list")
parser.add_argument("-id_list", '-ids',
    help="the file containing the IDs in the appropriate order")
parser.add_argument("-col_ids", 
    help="the file containing column IDs")
parser.add_argument("-rho_cutoff", '-rho',
    help="the cutoff to use for rho values",
    type = float)
parser.add_argument("-hdf5",
    action = 'store_true',
    default = False, 
    help="if the input file is in the hdf5 format")
parser.add_argument("-block_size",
    help = 'how many variables will be used at once for correlation analyzis; this helps keep memory requirements down and prevents a segmentation fault memory error',
     type = int,
     default = 5000)
parser.add_argument("-time",
    action = 'store_true',
    default = False, 
    help="if you only want to test the speed of relationship detection")
parser.add_argument("-sc_clust",
    action = 'store_true',
    default = False, 
    help="if scRNAseq clustering is going to be done")
parser.add_argument("-transpose",
    action = 'store_true',
    default = False, 
    help="if we're going to make an adjacency list on the columns rather than the rows")
parser.add_argument(
    "-rand_seed",'-seed',
    help='the random number input for random number seed, default = 12345',
    dest = 'rand_seed',
    type = int,
    default = 12345)

parser.add_argument("-hdf5_out",
    action = 'store_true',
    default = False, 
    help="if we want to make an hdf5 as the output instead of the rho dicts")
parser.add_argument("-euclidean_dist",
    action = 'store_true',
    default = False, 
    help="if we are going to also calculate the euclidean distance of all points against each other and store as an hdf5.")
parser.add_argument("-row_subset",
    help="a file with the row indices to include if we only want to do this on a subset of the rows. Note that this operation is performed before the transpose if need be.")
parser.add_argument(
    "-rho_dict_dir",
    help='the destination directory')

parser.add_argument(
    "-FPR_multiple",'-mult',
    help="the multiple to use for FPR based anti-correlation feature selection. Only use this if you really know what you're doing.",
    type = int,
    default = 15)
parser.add_argument(
    "-num_pos_cor_cutoff",
    help="the The minimum number of positive correlations that a gene must have to be included in the clustering. Default is 10. Only use this if you really know what you're doing.",
    type = int,
    default = 10)


parser.add_argument("-prop","-proportaionality",
    action = 'store_true',
    default = False, 
    help="if you only want to use proportionality instead of Spearman")

args = parser.parse_args()


np.random.seed(args.rand_seed)
random.seed(args.rand_seed)



## constants
FPR = 1000 ## 1 in 1000 false positive rate for negative correlation clustering
FPR_multiple = args.FPR_multiple ## the number of times the expected number of false positives to be included in clustering
pos_FPR = int(1e3)


##############################################################
args.infile_path = os.path.realpath(args.infile_path)

rho_cutoff_bool = True
rho_cutoff = args.rho_cutoff





##############

## run spearman analysis

infile_path=args.infile_path
if args.hdf5:
    h5f = h5py.File(infile_path, 'r')
    infile=h5f["infile"]
else:
	infile = np.array(read_table(infile_path))



if args.id_list is None and not args.hdf5:
    IDlist = list(infile[1:,0])
    col_ids = list(infile[0,1:])
    infile = np.array(infile[1:,1:],dtype = float)
elif args.id_list != None and not args.hdf5:
    infile = np.array(infile[1:,1:],dtype = float)
    ID_list = args.id_list
    IDlist = read_file(ID_list,'lines')
else:
    ID_list = args.id_list
    IDlist = read_file(ID_list,'lines')

## subset for the pertinent rows
if args.row_subset!=None:
    row_indices = read_file(args.row_subset,'lines')
    row_indices = sorted(list(map(int,row_indices)))
    infile = infile[row_indices,:]



## here's where we'll transpose the input
if args.transpose:
    if args.hdf5:
        infile = np.transpose(infile)
        IDlist = read_file(args.col_ids,'lines')[1:]# start at 1 because of the leader column
    else:
        infile = np.transpose(infile)
        IDlist = col_ids


num_vars = len(IDlist)
## sanity check that the ID list length is equal to the row num
if num_vars != np.shape(infile)[0]:
    print('the length of the ID list does not equal the number of rows in the input dataset')
    print("infile:",np.shape(infile)[0], )
    print("IDlist:",num_vars, IDlist[0:5])
    sys.exit()

index=0

temp_dir=str(infile_path).split('/')
temp_dir=('/').join(temp_dir[:-1])

dump_length = 20
bin_size = args.block_size
bin_size = min([bin_size,len(IDlist)])## this prevents the program from hitting index errors for small datasets


count=0


print('finding the spearman correlations for very big file')
print('this may take a while if your dataset has lots of variables...')

################################################
if args.rho_dict_dir==None:
    rho_dict_dir = temp_dir+'/rho_dicts'
else:
    rho_dict_dir = args.rho_dict_dir
cmd('mkdir '+rho_dict_dir)
#########################################################

#output_lines = [["X_var","Y_var","Spearman_p_val","Spearman_rho"]]
output_lines = []


total_vars = len(IDlist)
bins = []
cur_bin = 0
while cur_bin<total_vars:
	bins.append(min(cur_bin, total_vars))
	cur_bin+=bin_size

bins.append(total_vars)
if args.hdf5_out:
    print('making the hdf5 spearman output file')
    ## make the hdf5 output file
    hdf5_spear_out_file = rho_dict_dir+"/spearman.hdf5"
    print(hdf5_spear_out_file)
    #cmd('rm '+hdf5_spear_out_file)
    
    spear_f = h5py.File(hdf5_spear_out_file, "a")
    try:
        spear_out_hdf5 = spear_f.create_dataset("infile", (total_vars,total_vars), dtype=np.float16)
    except:
        spear_out_hdf5 = spear_f["infile"]



print(np.shape(infile))


print(infile)

print(bins)


start = time.time()
print("finding block...")
for i in range(0,(len(bins)-1)):
    for j in range(i,(len(bins)-1)):
        if (i!=j) or (len(bins) == 2):#True:
            out_rho_dict = rho_dict_dir+"/rho_"+str(bins[i])+"_to_"+str(bins[i+1])+"_vs_"+str(bins[j])+"_to_"+str(bins[j+1])+".pkl"

            ## if we're writing to an hdf5 file, check if the values are already entered
            if args.hdf5_out:
                #do_top_left = np.all(spear_out_hdf5[bins[i]:bins[i+1],bins[i]:bins[i+1]]==0)
                #do_top_right = np.all(spear_out_hdf5[bins[i]:bins[i+1],bins[j]:bins[j+1]]==0)
                #do_bottom_left = np.all(spear_out_hdf5[bins[j]:bins[j+1],bins[i]:bins[i+1]]==0)
                #do_bottom_right = np.all(spear_out_hdf5[bins[j]:bins[j+1],bins[j]:bins[j+1]]==0)
                do_top_left = np.all(spear_out_hdf5[bins[i],bins[i]:bins[i+1]]==0)
                do_top_right = np.all(spear_out_hdf5[bins[i],bins[j]:bins[j+1]]==0)
                do_bottom_left = np.all(spear_out_hdf5[bins[j],bins[i]:bins[i+1]]==0)
                do_bottom_right = np.all(spear_out_hdf5[bins[j],bins[j]:bins[j+1]]==0)
                do_one = (do_top_left or do_top_right or do_bottom_left or do_bottom_right)
                print(do_top_left , do_top_right , do_bottom_left , do_bottom_right)
                #print(spear_out_hdf5[bins[j]:bins[j+1],bins[j]:bins[j+1]])
            else:
                do_one = False
            
            #print((not os.path.exists(out_rho_dict) and not args.hdf5_out) , (args.hdf5_out and do_one))
            #sys.exit()
            if (not os.path.exists(out_rho_dict) and not args.hdf5_out) or (args.hdf5_out and do_one):
                ## find the spearman correlations

                print('working on',bins[i],bins[i+1],'vs',bins[j],bins[j+1])
                r=no_p_spear(infile[bins[i]:bins[i+1],],infile[bins[j]:bins[j+1],], axis = 1, prop=args.prop)
                
                ## if we're not making the spearman correlation matrix into an hdf5 file - save the dictionary:
                if not args.hdf5_out:
                    save_dict(r,out_rho_dict)

                ## if we are making the spearman correlation matrix an hdf5 file, we've got some work to do:
                else:
                    print(np.shape(r))
                    if np.shape(r)[0] == 2*bin_size:
                        #print('is symmetric:',r[bin_size:,bin_size:] == r[:bin_size,:bin_size])
                        if len(bins)==2:#False#np.shape(r)[0]<=bin_size:
                            ## there is a special case, where the number of variables is less than the
                            ## bin size. In this case, the returned matrix will be symmetric in all 4
                            ## quadrants, so we'll reduce it down to just the one
                            r = r[bin_size:,bin_size:]
                    print(np.shape(r))


                    ## because h5py's fancy indexing doesn't work very well with slicing in 2D...

                    ## first set the top left corner bins
                    if np.all(spear_out_hdf5[bins[i]:bins[i+1],bins[i]:bins[i+1]]==0):
                        print("setting top left")
                        spear_out_hdf5[bins[i]:bins[i+1],bins[i]:bins[i+1]] = np.array(r[:bin_size,:bin_size],dtype=np.float16)
                    
                    ## top right
                    if len(bins)!=2:
                        try:
                            if np.all(spear_out_hdf5[bins[i]:bins[i+1],bins[j]:bins[j+1]]==0):
                                print("setting top right")
                                spear_out_hdf5[bins[i]:bins[i+1],bins[j]:bins[j+1]] = np.array(r[:bin_size,bin_size:],dtype=np.float16)
                        except:
                            print(sys.exit('error setting the top right'))
                            pass
                    
                    ## bottom left
                    if len(bins)!=2:
                        try:
                            if np.all(spear_out_hdf5[bins[j]:bins[j+1],bins[i]:bins[i+1]]==0):
                                print('setting bottom left')
                                spear_out_hdf5[bins[j]:bins[j+1],bins[i]:bins[i+1],] = np.array(r[bin_size:,:bin_size],dtype=np.float16)
                        except:
                            print(sys.exit('error setting the bottom left'))
                            pass
                    
                    ## bottom right
                    if len(bins)!=2:
                        try:
                            if np.all(spear_out_hdf5[bins[j]:bins[j+1],bins[j]:bins[j+1]]==0):
                                print('setting bottom right')
                                spear_out_hdf5[bins[j]:bins[j+1],bins[j]:bins[j+1]] = np.array(r[bin_size:,bin_size:],dtype=np.float16)
                        except:
                            print(sys.exit('error setting the bottom right'))
                            pass
            else:
                print('already finished block',bins[i],bins[i+1],'vs',bins[j],bins[j+1])

end = time.time()



if args.euclidean_dist:
    from sklearn import metrics
    euclidean_distances = metrics.pairwise.euclidean_distances
    print('making the negative euclidean distance matrix')
    ## make the euclidean distance output matrix
    ## make the hdf5 output file
    hdf5_euc_out_file = rho_dict_dir+"/neg_euc_dist.hdf5"
    print(hdf5_euc_out_file)
    euc_f = h5py.File(hdf5_euc_out_file, "a")
    ## set up the data matrix (this assumes float32)
    float_type = np.float16
    try:
        neg_euc_out_hdf5 = euc_f.create_dataset("infile", (total_vars,total_vars), dtype=float_type)
    except:
        neg_euc_out_hdf5 = euc_f["infile"]
    # else:
    #     neg_euc_out_hdf5 = euc_f.create_dataset("infile", (total_vars,total_vars), dtype=np.float32)
    # ## go through and calculate the negative euclidean distances
    ###########################
    spear_min = 99999
    temp_start=int(0)
    temp_end=int(args.block_size)
    print("calculating the minimum non-zero Spearman rho")
    while temp_end < spear_out_hdf5.shape[0]:
        print(round(100*temp_start/spear_out_hdf5.shape[0],2),"% ","done")
        spear_min=min([spear_min,np.nanmin(spear_out_hdf5[temp_start:temp_end,:])/np.log2(total_vars)])
        temp_start+=args.block_size
        temp_end+=args.block_size
    ## catch the last bit
    spear_min=min([spear_min,np.nanmin(spear_out_hdf5[temp_start:,:])])
    #spear_min = np.nanmin(spear_out_hdf5)/np.log2(total_vars)
    print("spear_min",spear_min)
    ###########################
    for i in range(0,(len(bins)-1)):
        for j in range(i,(len(bins)-1)):
            #print(np.all(neg_euc_out_hdf5[bins[i]:bins[i+1],bins[j]:bins[j+1]]==0) , np.all(neg_euc_out_hdf5[bins[j]:bins[j+1],bins[i]:bins[i+1]]==0))
            if (np.all(neg_euc_out_hdf5[bins[i],bins[j]:bins[j+1]]==0) and np.all(neg_euc_out_hdf5[bins[j],bins[i]:bins[i+1]]==0)):
                print('calculating negative euclidean distance for',bins[i],bins[i+1],'vs',bins[j],bins[j+1])
                #temp_neg_euc = -euclidean_distances(np.array(spear_out_hdf5[bins[i]:bins[i+1],:],dtype=np.float32),np.array(spear_out_hdf5[bins[j]:bins[j+1],:],dtype=np.float32),squared=True)
                #temp_neg_euc = -euclidean_distances(np.array(spear_out_hdf5[bins[i]:bins[i+1],:],dtype=float_type),np.array(spear_out_hdf5[bins[j]:bins[j+1],:],dtype=float_type),squared=True)/np.log2(total_vars)
                temp_subset_mat_1 = np.array(spear_out_hdf5[bins[i]:bins[i+1],:],dtype=float_type)/np.log2(total_vars)
                temp_subset_mat_2 = np.array(spear_out_hdf5[bins[j]:bins[j+1],:],dtype=float_type)/np.log2(total_vars)
                #print('before')
                #print(temp_subset_mat_1[np.where(np.isnan(temp_subset_mat_1))])
                temp_subset_mat_1[np.where(np.isnan(temp_subset_mat_1))]=spear_min
                #print(temp_subset_mat_1[np.where(np.isnan(temp_subset_mat_1))])
                #print('after')
                temp_subset_mat_2[np.where(np.isnan(temp_subset_mat_2))]=spear_min
                #print(temp_subset_mat_1)
                #print(temp_subset_mat_2)
                num_nan_1 = np.sum(np.isnan(temp_subset_mat_1))
                num_inf_1 = np.sum(np.isinf(temp_subset_mat_1))
                num_nan_2 = np.sum(np.isnan(temp_subset_mat_2))
                num_inf_2 = np.sum(np.isinf(temp_subset_mat_2))
                if num_nan_1+num_inf_1+num_nan_2+num_inf_2 >0:
                    print('num_nan_1:',num_nan_1)
                    print('num_inf_1:',num_inf_1)
                    print('num_nan_2:',num_nan_2)
                    print('num_inf_2:',num_inf_2)
                else:
                    pass#print('no nans or infs')
                temp_neg_euc = -euclidean_distances(temp_subset_mat_1,temp_subset_mat_2)*np.log2(total_vars)
                neg_euc_out_hdf5[bins[i]:bins[i+1],bins[j]:bins[j+1]] = temp_neg_euc
                neg_euc_out_hdf5[bins[j]:bins[j+1],bins[i]:bins[i+1]] = np.transpose(temp_neg_euc)
                #print(temp_neg_euc)
                #print(neg_euc_out_hdf5[:,:])
                #neg_euc_out_hdf5 = np.transpose(temp_neg_euc)
                #print(np.all(neg_euc_out_hdf5[bins[i]:bins[i+1],bins[j]:bins[j+1]]==0) , np.all(neg_euc_out_hdf5[bins[j]:bins[j+1],bins[i]:bins[i+1]]==0))
            else:
                print('already finished',bins[i],bins[i+1],'vs',bins[j],bins[j+1])
    for i in range(0,np.shape(neg_euc_out_hdf5)[0]):
        neg_euc_out_hdf5[i,i]=-0.0

    euc_f.close()



if args.hdf5_out:
    ## close the hdf5 file
    spear_f.close()


if args.time:
    print((end-start)/60,'minutes spent finding correlations')
    sys.exit()


###############################################################



################################################################################
## read in the ID list
print('logging all the IDs')

ID_hash = {}
for i in range(0,len(IDlist)):
	ID_hash[IDlist[i]] = i
	

################################################################################
## get the rho dictionaries

## parse the name of the file to get the 
def parse_dict_name(dict_path):
	dict_path = dict_path.split('/')
	dict_path = dict_path[-1]
	dict_path = dict_path.strip('rho_')
	dict_path = dict_path.strip('.pkl')
	dict_path = dict_path.split('_vs_')
	dict_path[0] = dict_path[0].split('_to_')
	dict_path[1] = dict_path[1].split('_to_')
	dict_path[0][0]=int(dict_path[0][0])
	dict_path[0][1]=int(dict_path[0][1])
	dict_path[1][0]=int(dict_path[1][0])
	dict_path[1][1]=int(dict_path[1][1])
	temp_id_indices = np.concatenate((np.arange(dict_path[0][0],dict_path[0][1]),np.arange(dict_path[1][0],dict_path[1][1])))
	temp_IDs=[IDlist[i] for i in temp_id_indices]
	return((temp_id_indices,temp_IDs))

rho_properties_dict = {}
rho_files = []
for glob in glob.glob(rho_dict_dir+'/*.pkl'):
	rho_files.append(glob)
	rho_properties_dict[glob]=parse_dict_name(glob)
	#print(glob,'\n',rho_properties_dict[glob])

	
################################################################################
## make the adjacency list
print('making the adjacency list')


#################
## look at the sum of all negative correlations
sum_neg = []
all_vars_adj_dict = {}
neg_vars_adj_dict = {}
total_neg_vars_adj_dict = {}
for i in range(0,len(IDlist)):
    ## [ID, sum_negative_rho, count_below_threshold, total_number_negative, (count_below_threshold+1)/(count_above_threshold +1)]
    sum_neg.append([IDlist[i],0,0,0,0])
    all_vars_adj_dict[IDlist[i]] = np.zeros(len(IDlist,),dtype = bool)
    neg_vars_adj_dict[IDlist[i]] = np.zeros(len(IDlist,),dtype = bool)
    total_neg_vars_adj_dict[IDlist[i]] = np.zeros(len(IDlist,),dtype = bool)
#################

####
def get_density(in_vect):
    in_vect = np.array(in_vect)
    in_vect = in_vect[np.array((np.isnan(in_vect)*-1)+1,dtype=bool)]## remove nans
    in_vect = in_vect[np.array((np.isinf(in_vect)*-1)+1,dtype=bool)]## remove infs
    offset = min(in_vect)
    #print(offset)
    in_vect = in_vect - offset
    #print(in_vect)
    density = gaussian_kde(in_vect)
    #print(in_vect)
    #print(density)
    xs = np.arange(0,max(in_vect),max(in_vect)/150)
    #print(xs)
    #density.covariance_factor = lambda : 2.5
    #density._compute_covariance()
    #print(density(xs))
    y = density(xs)
    y = y/sum(y)
    return(xs+offset,y)

def plot_one_density(neg_cor_density_x, neg_cor_density_y,out_file, label = 'negative correlation Rhos from bootstrap shuffled negative control'):
    global args
    x_min = np.min(neg_cor_density_x)
    y_min = np.min(neg_cor_density_y)
    x_max = np.max(neg_cor_density_x)
    y_max = np.max(neg_cor_density_y)

    plt.clf()
    plt.plot(neg_cor_density_x, neg_cor_density_y, 
        label = label,
        color = 'blue',
        linewidth = 3)
    plt.legend()
    plt.savefig(out_file,
        dpi=600,
        bbox_inches='tight')

def plot_one_density_x_cutoff(neg_cor_density_x, neg_cor_density_y,cutoff,out_file,log = False,poisson_mu=None,label = "density"):
    global args
    from scipy.stats import poisson
    x_min = np.min(neg_cor_density_x)
    y_min = np.min(neg_cor_density_y)
    x_max = np.max(neg_cor_density_x)
    y_max = np.max(neg_cor_density_y)

    plt.clf()
    plt.plot(neg_cor_density_x, neg_cor_density_y, 
        label = label,
        color = 'blue',
        linewidth = 3)
    plt.plot([cutoff,cutoff],[y_min,y_max],
        color = 'k',
        lw = 2,
        label = 'cutoff')

    ## calculate the poisson at the given mu=cutoff
    if poisson_mu != None:
        temp_x = np.arange(x_max-x_min)+x_min
        dist = poisson(poisson_mu)
        plt.plot(np.log2(temp_x+1), np.log2(dist.pmf(temp_x)+1), ls='--', color='red',
                     label='poisson' , linestyle='steps-mid')
    plt.legend()

    plt.savefig(out_file,
        dpi=600,
        bbox_inches='tight')


def plot_densities(neg_cor_density_x, neg_cor_density_y,series_2_x,series_2_y,out_file):
    global args
    x_min = min([np.min(neg_cor_density_x),np.min(series_2_x)])
    y_min = min([np.min(neg_cor_density_y),np.min(series_2_y)])
    x_max = max([np.max(neg_cor_density_x),np.max(series_2_x)])
    y_max = max([np.max(neg_cor_density_y),np.max(series_2_y)])

    plt.clf()
    plt.plot(neg_cor_density_x, neg_cor_density_y, 
        label = 'sum of negative correlations per gene',
        color = 'blue',
        linewidth = 3)
    plt.plot(series_2_x, series_2_y, 
        label = 'bootstrap shuffled negative control',
        color = 'red',
        linewidth = 3)
    plt.legend()
    plt.savefig(out_file,
        dpi=600,
        bbox_inches='tight')

def plot_avg_densities(neg_cor_density_x, neg_cor_density_y,series_2_x,series_2_y,out_file):
    global args
    x_min = min([np.min(neg_cor_density_x),np.min(series_2_x)])
    y_min = min([np.min(neg_cor_density_y),np.min(series_2_y)])
    x_max = max([np.max(neg_cor_density_x),np.max(series_2_x)])
    y_max = max([np.max(neg_cor_density_y),np.max(series_2_y)])

    plt.clf()
    plt.plot(neg_cor_density_x, neg_cor_density_y, 
        label = 'average Rho of negative correlations per gene',
        color = 'blue',
        linewidth = 3)
    plt.plot(series_2_x, series_2_y, 
        label = 'average Rho of negative correlations per boostrap negative control',
        color = 'red',
        linewidth = 3)
    plt.legend()
    plt.savefig(out_file,
        dpi=600,
        bbox_inches='tight')

def density_scatter(x,y,out_file,log_density=0,expected_x=[],expected_y=[]):
    ## remove points that are on the origin
    x=np.array(x)
    y=np.array(y)
    non_origin = (np.array(x==0, dtype = int) + np.array(y==0,dtype=int)) < 2
    x = x[non_origin].tolist()
    y = y[non_origin].tolist()
    if len(x)<2:
        return()
    xy = np.vstack([x,y])
    try:
        z = gaussian_kde(xy)(xy)
    except:
        print("error in getting the gaussian KDE")
        return()
    print(z)
    if log_density:
        while log_density:
            log_density-=1
            z = np.log2(z+1)
            z = z-min(z)
            z = z/max(z)
            print(z)
    plt.clf()
    fig, ax = plt.subplots()
    ax.scatter(x, y, c=z, s=5, edgecolor='', cmap=plt.cm.jet)
    if expected_x != [] and expected_y != [] :
        ax.plot(expected_x,expected_y,linestyle='-',c='black')
        plt.xlim((-1,max(x)+1))
        plt.ylim((0,max(y)+1))

    plt.savefig(out_file,
        dpi=600,
        bbox_inches='tight'
        )


###################
small_adj_list = []

do_sum_neg = args.sc_clust

def calculate_mad(in_vect, dev_num = 3, mean = False):
    med = np.median(in_vect)
    if mean:
        dev = np.mean(np.abs(in_vect - med))
    else:
        dev = np.median(np.abs(in_vect - med))
    print('median:',med)
    print('median absolute deviation:',dev)
    cutoff = med + (dev_num * dev)
    print('cutoff =',cutoff)
    return(med,dev,cutoff)

# def get_Z_cutoff(vect,z=4.5,positive = False):
#     ## this is for a half distribution, so mult by -1 to make it 'normal'
#     global FPR
#     vect_original = np.array(deepcopy(vect))
#     vect = list(vect)
#     vect += list(np.array(vect)*-1)
#     vect = np.array(vect)
#     ## get the mean, & SD
#     v_mean = np.mean(vect)
#     v_std = np.std(vect)
#     cutoff = v_std * z
#     if not positive:
#         cutoff *= -1
#         cutoff = max([-0.9,cutoff])
#     else:
#         cutoff = min([0.9,cutoff])
#     print('cutoff:',cutoff)
#     if not positive:
#         num_sig = np.sum(np.array(vect_original<cutoff,dtype = bool))
#         total_num = vect_original.shape[0]
#         print(num_sig, "/", total_num, 'significant')
#         FPR = max([1,total_num])/num_sig
#         return(cutoff, FPR)
#     else:
#         return(cutoff)


# def get_FPR_cutoff(negative_control_negative_rho_vect,FPR,positive = False):
#     ####################################################################
#     ## figure out the cutoff to use for a given false positive rate
#     negative_control_negative_rho_vect = negative_control_negative_rho_vect[np.where(abs(negative_control_negative_rho_vect)!=1)[0]]
#     negative_control_negative_rho_vect = np.sort(negative_control_negative_rho_vect, kind = "mergesort")
#     if positive:
#         negative_control_negative_rho_vect = negative_control_negative_rho_vect[::-1]
#     print("sorting list:")
#     print(negative_control_negative_rho_vect[:5],'...',negative_control_negative_rho_vect[-5:])
#     number_of_total_negative_correlations_in_boot = np.shape(negative_control_negative_rho_vect)[0]
#     print("number_of_total_negative_correlations_in_boot")
#     print(number_of_total_negative_correlations_in_boot)
#     print("FPR:",FPR)

#     neg_cutoff_index = int(max([1,number_of_total_negative_correlations_in_boot/FPR])-1)## this is an emperical cutoff of 1/FPR randomly negative rhos
#     print("cutoff_index")
#     print(neg_cutoff_index)

#     neg_cutoff = float(negative_control_negative_rho_vect[neg_cutoff_index])
#     #print(negative_control_negative_rho_vect[:neg_cutoff_index])
#     print('Rho cutoff:',neg_cutoff)
#     return(neg_cutoff)

    ##################################################################



if do_sum_neg or args.rho_cutoff==None:
    ## run the bootstrap negative control for finding sum negative relationships
    ## first get a bin_size random set of variables to do the negative control on
    neg_control_sample_idxs = np.arange(len(IDlist))
    ## remove any genes that are zero in all samples
    non_zero_idxs = np.where(np.sum(infile,axis=1)>0)[0]
    neg_control_sample_idxs = neg_control_sample_idxs[non_zero_idxs]
    np.random.shuffle(neg_control_sample_idxs)
    print(np.shape(neg_control_sample_idxs))
    bin_size = min([bin_size, np.shape(neg_control_sample_idxs)[0]])
    print(bin_size)
    neg_control_sample_idxs = neg_control_sample_idxs[:bin_size]
    neg_control_sample_idxs = neg_control_sample_idxs.tolist()
    neg_control_sample_idxs.sort()
    print(np.shape(neg_control_sample_idxs))

    ## subset these IDs
    neg_control_subset_mat = np.array(infile[neg_control_sample_idxs,:])
    print('making negative control bootstrap shuffled matrix',np.shape(neg_control_subset_mat))

    ## go through each row and shuffle them within variables
    for i in range(0,np.shape(neg_control_subset_mat)[0]):
        temp_row_shuffle_order = np.arange(np.shape(neg_control_subset_mat)[1])
        np.random.shuffle(temp_row_shuffle_order)
        temp_row_shuffle_order = temp_row_shuffle_order.tolist()

        neg_control_subset_mat[i,:] = neg_control_subset_mat[i,temp_row_shuffle_order]

    ## get the spearman_rhos for the shuffled matrix
    r=no_p_spear(neg_control_subset_mat,neg_control_subset_mat, axis = 1, prop=args.prop)
    r = r[:bin_size,:bin_size]
    print("rho shape",np.shape(r))
    ## remove self correlations
    upper_tiangle_indices = np.triu_indices(bin_size,k=1)
    
    
    ## subset the negative ones
    print(r)
    linear_rho = r[upper_tiangle_indices]
    print(linear_rho)
    print(max(linear_rho))
    print(len(np.where(linear_rho == max(linear_rho))[0]))
    ## calculate the median absolute deviation
    #calculate_mad(linear_rho)

    negative_control_negative_rho_vect = linear_rho[np.where(linear_rho<0)[0]]
    negative_control_positive_rho_vect = linear_rho[np.where(linear_rho>0)[0]]
    #calculate_mad(negative_control_negative_rho_vect)

    ## plot the average Rho per negative correlation
    print(negative_control_negative_rho_vect)
    neg_cor_density_x, neg_cor_density_y = get_density(negative_control_negative_rho_vect)
    pos_cor_density_x, pos_cor_density_y = get_density(negative_control_positive_rho_vect)
    
    
    sum_neg_plot = os.path.dirname(args.adj_list_out)+'/boostrap_neg_cor_rhos.png'
    plot_one_density(neg_cor_density_x, neg_cor_density_y, sum_neg_plot, label="density of negative correlation null distribution")
    pos_rho_plot = os.path.dirname(args.adj_list_out)+'/boostrap_pos_cor_rhos.png'
    plot_one_density(pos_cor_density_x, pos_cor_density_y,pos_rho_plot, label="density of positive correlation null distribution")

    ## plot the distribution of all bootstrap negative control Rhos
    neg_cor_density_x, neg_cor_density_y = get_density(linear_rho)
    sum_neg_plot = os.path.dirname(args.adj_list_out)+'/boostrap_cor_rhos.png'
    plot_one_density(neg_cor_density_x, neg_cor_density_y, sum_neg_plot,label="density of all correlation null distribution")

    print("\n\ncalculating negative correlation FPR cutoff from null distribution:")
    #neg_cutoff = get_FPR_cutoff(negative_control_negative_rho_vect,FPR)
    neg_cutoff, FPR = get_Z_cutoff(negative_control_negative_rho_vect)
    print('\nEmpiric FPR:',FPR)
    print("\n\ncalculating positive correlation FPR cutoff from null distribution:")
    #pos_cutoff = get_FPR_cutoff(negative_control_positive_rho_vect,pos_FPR, positive = True)
    pos_cutoff = get_Z_cutoff(negative_control_positive_rho_vect, z=5.5, positive = True)
    rho_cutoff = pos_cutoff
    args.rho_cutoff = pos_cutoff
    #sys.exit()
#################################################################################
    
## make the output file
cmd('rm '+args.adj_list_out)
cmd('rm '+args.adj_list_out[:-4]+"_dedup.tsv")
out_file_name = args.adj_list_out[:-4]+"_dedup.tsv"
out_file = open(out_file_name, "a")

#out_file.write('\t'.join(['var_1', 'var_2', 'NA', 'NA','NA','NA','rho'])+'\n')
if not args.prop:
    out_file.write('\t'.join(['var_1', 'var_2', 'rho'])+'\n')
else:
    out_file.write('\t'.join(['var_1', 'var_2', 'phi_proportionality'])+'\n')


for rho in rho_files:
    num_pos_cor_dict = {str(temp_id):0 for idx, temp_id in enumerate(IDlist)}
    ## go through all the rho dict files
    print('reading in', rho)
    temp_rho_mat = import_dict(rho)
    temp_indices, temp_IDs = rho_properties_dict[rho]
    for i in range(0,np.shape(temp_rho_mat)[0]):
        ## go through each line in the rho matrix and log the pertinent indices
        cur_i_variable = temp_IDs[i]
        sig_indices = np.where(np.absolute(temp_rho_mat[i,:])>=args.rho_cutoff)[0]
        sig_IDs = [temp_IDs[j] for j in sig_indices]
        sig_rhos = temp_rho_mat[i,sig_indices]

        ## go through the significant indicies
        for q in range(0,len(sig_IDs)):
            if sig_IDs[q] != cur_i_variable:
                if ID_hash[cur_i_variable] > ID_hash[sig_IDs[q]]:
                    ## check if variable pair has been done already
                    temp_relationships = all_vars_adj_dict[cur_i_variable]
                    var2_idx = ID_hash[sig_IDs[q]]
                    ## check if it has been done (ie: currently set to zero.)
                    if temp_relationships[var2_idx] == 0:
                        temp_relationships[var2_idx]+=1## this is the indicator that we've doen it already
                        all_vars_adj_dict[cur_i_variable] = temp_relationships
                        #temp_out_line = [str(cur_i_variable), str(sig_IDs[q]), 'NA', 'NA','NA','NA',str(sig_rhos[q])]
                        temp_out_line = [str(cur_i_variable), str(sig_IDs[q]), str(sig_rhos[q])]
                        out_file.write('\t'.join(temp_out_line)+'\n')
                        ## also log the positive correlations
                        num_pos_cor_dict[str(cur_i_variable)] = 1 + num_pos_cor_dict[str(cur_i_variable)] 
                        num_pos_cor_dict[str(sig_IDs[q])] = 1 + num_pos_cor_dict[str(sig_IDs[q])] 
        
        if do_sum_neg:
            ## do the same for all of the negative correlations, regardless of magnitude
            total_negative_indices = np.where(temp_rho_mat[i,:]<0)[0]
            neg_IDs = [temp_IDs[j] for j in total_negative_indices]
            neg_rhos = temp_rho_mat[i,total_negative_indices]
            for q in range(0,len(neg_rhos)):
                if ID_hash[cur_i_variable] > ID_hash[neg_IDs[q]]:

                    ## check if variable pair has been done already
                    temp_relationships = total_neg_vars_adj_dict[cur_i_variable]
                    var2_idx = ID_hash[neg_IDs[q]]
                    if temp_relationships[var2_idx] == 0:
                        ## total number of Rho <0
                        temp_relationships[var2_idx]+=1
                        total_neg_vars_adj_dict[cur_i_variable] = temp_relationships
                        sum_neg[ID_hash[cur_i_variable]][3]+=1
                        sum_neg[ID_hash[neg_IDs[q]]][3]+=1

                        ## then if it's actually less than the simulation determined cutoff, log it for that as well
                        if neg_rhos[q]<neg_cutoff:# and neg_rhos[q]>-1 :
                            sum_neg[ID_hash[cur_i_variable]][1]+=neg_rhos[q]
                            sum_neg[ID_hash[neg_IDs[q]]][1]+=neg_rhos[q]

                            ## total number < auto-determined Rho cutoff from bootstrap shuffled negative correlations
                            sum_neg[ID_hash[cur_i_variable]][2]+=1
                            sum_neg[ID_hash[neg_IDs[q]]][2]+=1


out_file.close()

if do_sum_neg:
    ## sum up the negative correlations for the bootstrap shuffled negative control
    negative_control_sum_neg = []
    negative_control_count_neg = []
    for i in range(0,np.shape(r)[0]):
        neg_indices = np.where(r[i,:]<0)[0]
        count_neg = np.shape(neg_indices)[0]
        negative_control_count_neg.append(count_neg)
        sum_neg_rhos = sum(r[i,neg_indices])
        negative_control_sum_neg.append(sum_neg_rhos)

    ## calculate the average per comparison for each gene, then multiply it by the number of comparisons 
    ## in the full dataset. This extrapolates the subset to the size of the full dataset.
    ## this will then be used to dynamicaly set the cutoff of what counts as a negative correlation 
    ## at the desired FDR (default FDR = 0.05)
    negative_control_avg_neg = np.array(negative_control_sum_neg)/np.array(negative_control_count_neg)
    negative_control_sum_neg = (np.array(negative_control_sum_neg)/bin_size)*len(IDlist)
    #negative_control_avg_neg = 
    max_neg_ctr = max(negative_control_sum_neg)
    print("max of the negative control =",max_neg_ctr)


    ######################################################################################
    ## process the sum negative table into a vector so that we can do interesting things with it
    print(sum_neg[:10])
    sum_neg_vect = np.zeros(len(sum_neg),dtype = float)
    sum_neg_count = np.zeros(len(sum_neg),dtype = float)
    sum_neg_count_total = np.zeros(len(sum_neg),dtype = float)
    non_sig_neg_count = np.zeros(len(sum_neg),dtype = float)
    for i in range(0,len(sum_neg)):
        sum_neg_vect[i]=sum_neg[i][1]
        sum_neg_count[i]=sum_neg[i][2]
        sum_neg_count_total[i]=sum_neg[i][3]
        non_sig_neg_count[i] = sum_neg[i][3] - sum_neg[i][2]

    sum_neg_vect = np.array(sum_neg_vect)
    sum_neg_count = np.array(sum_neg_count)
    sum_neg_count_total = np.array(sum_neg_count_total)
    count_above_threshold = sum_neg_count_total - sum_neg_count
    count_above_below_ratio = (sum_neg_count+1)/(count_above_threshold+1)

    ## log the ratios in the output file
    for i in range(0,len(sum_neg)):
        sum_neg[i][4]=count_above_below_ratio[i]

    avg_neg = sum_neg_vect/sum_neg_count
    print(sum_neg_vect[:10])


    ## neg count
    #print('\n\t\tcalculating median absolute deviation for the total significant negative count')
    # given the total number of observed negative correlations (expected_y)
    print("max of sum_neg_count_total",np.max(sum_neg_count_total))
    expected_y = np.arange(np.max(sum_neg_count_total))
    expected_x = expected_y/(FPR/FPR_multiple)
    expected_x = np.log2(expected_x)
    expected_x = expected_x[1:]
    expected_y = expected_y[1:]
    print(expected_y)
    print(expected_x)
    #sys.exit()
    sum_neg_count = np.log2(sum_neg_count+1)
    #med,med_dev,cutoff = calculate_mad(sum_neg_count,mean=False)

    ## append whether or not they passed the cutoff to the summary matrix
    def pass_cutoff(above, total, num_pos_cor, num_pos_cor_cutoff = 10):
        global FPR,FPR_multiple, expected_x, expected_y
        ## find the nearest for total (which corresponds to expected_y)
        lowest_index = np.argmin(abs(total - expected_y))
        ## the corresponding y index is the cutoff value
        temp_cutoff = expected_x[lowest_index]
        # if above < 10:
        #     return("False")
        # elif above > temp_cutoff:
        #     return("True")
        # else:
        #     return("False")
        if above < 1:#10:
            return("False")
        elif total/above < (FPR/FPR_multiple) and num_pos_cor >= num_pos_cor_cutoff:
            return("True")
        else:
            return("False")

    #[ID, sum_negative_rho, count_below_threshold, total_number_negative, (count_below_threshold+1)/(count_above_threshold +1)]
    included_indices = []
    for i in range(0,len(sum_neg)):
        sum_neg[i].append(pass_cutoff(sum_neg[i][2],sum_neg[i][3], num_pos_cor_dict[sum_neg[i][0]], num_pos_cor_cutoff = args.num_pos_cor_cutoff))
        if eval(sum_neg[i][-1]):
            included_indices.append(i)
        # sum_neg[]
        # temp_pass = "False"
        # if sum_neg_count[i] > np.log2(expected_false_positives+1):
        #     temp_pass = "True"
        # sum_neg[i].append(temp_pass)


    # ## plot the average Rho per negative correlation
    # neg_cor_density_x, neg_cor_density_y = get_density(avg_neg)
    # series_2_x, series_2_y = get_density(negative_control_sum_neg)
    
    # sum_neg_plot = os.path.dirname(args.adj_list_out)+'/avg_neg_cor.png'
    # plot_densities(neg_cor_density_x, neg_cor_density_y, series_2_x, series_2_y, sum_neg_plot) 

    ## plot the ratio of negative correlations below and above the bootstrap shuffled determined cutoff
    print('count_above_below_ratio',count_above_below_ratio)
    neg_cor_density_x, neg_cor_density_y = get_density(count_above_below_ratio)
    
    sum_neg_plot = os.path.dirname(args.adj_list_out)+'/ratio_of_neg_cor_above_below_cutoff.png'
    plot_one_density(neg_cor_density_x, neg_cor_density_y, sum_neg_plot) 

    ## plot the counts of negative correlations against
    neg_cor_density_x, neg_cor_density_y = get_density(count_above_below_ratio)
    
    sum_neg_plot = os.path.dirname(args.adj_list_out)+'/ratio_of_neg_cor_above_below_cutoff.png'
    plot_one_density_x_cutoff(neg_cor_density_x, neg_cor_density_y, 1/1000, sum_neg_plot) 

    ## plot the counts of significant vs non-significant
    count_sig_vs_non_sig = os.path.dirname(args.adj_list_out)+'/sig_neg_count_vs_total_neg_count.png'
    density_scatter(sum_neg_count,sum_neg_count_total,count_sig_vs_non_sig,log_density=4,expected_x=expected_x,expected_y=expected_y)
    count_sig_vs_non_sig = os.path.dirname(args.adj_list_out)+'/sig_neg_count_vs_total_neg_count_included_only.png'
    if len(included_indices)>0:
        density_scatter(sum_neg_count[included_indices],sum_neg_count_total[included_indices],count_sig_vs_non_sig,log_density=4,expected_x=expected_x,expected_y=expected_y)



    ## write the results to file
    sum_neg_out = os.path.dirname(args.adj_list_out)+'/sum_neg_cor.txt'
    write_table(sum_neg,sum_neg_out)




################################################################################

#cmd('remove_dups.py -infile '+args.adj_list_out)















#!/usr/bin/env python3
##import dependency libraries
import sys,time,glob,os,pickle,fileinput,argparse
from subprocess import Popen
from operator import itemgetter
import gc, fileinput
import numpy as np
#import pandas as pd
##############################################################
## basic function library
def read_file(tempFile,linesOraw='lines',quiet=False):
    if not quiet:
        print('reading',tempFile)
    f=open(tempFile,'r')
    if linesOraw=='lines':
        lines=f.readlines()
        for i in range(0,len(lines)):
            lines[i]=lines[i].strip('\n')
    elif linesOraw=='raw':
        lines=f.read()
    f.close()
    return(lines)

def make_file(contents,path):
    f=open(path,'w')
    if isinstance(contents,list):
        f.writelines(contents)
    elif isinstance(contents,str):
        f.write(contents)
    f.close()

    
def flatten_2D_table(table,delim):
    #print(type(table))
    if str(type(table))=="<class 'numpy.ndarray'>":
        out=[]
        for i in range(0,len(table)):
            out.append([])
            for j in range(0,len(table[i])):
                try:
                    str(table[i][j])
                except:
                    print(table[i][j])
                else:
                    out[i].append(str(table[i][j]))
            out[i]=delim.join(out[i])+'\n'
        return(out)
    else:
        for i in range(0,len(table)):
            for j in range(0,len(table[i])):
                try:
                    str(table[i][j])
                except:
                    print(table[i][j])
                else:
                    table[i][j]=str(table[i][j])
            table[i]=delim.join(table[i])+'\n'
    #print(table[0])
        return(table)

def strip_split(line, delim = '\t'):
    return(line.strip('\n').split(delim))

def make_table(lines,delim):
    for i in range(0,len(lines)):
        lines[i]=lines[i].strip()
        lines[i]=lines[i].split(delim)
        for j in range(0,len(lines[i])):
            try:
                float(lines[i][j])
            except:
                lines[i][j]=lines[i][j].replace('"','')
            else:
                lines[i][j]=float(lines[i][j])
    return(lines)


def get_file_path(in_path):
    in_path = in_path.split('/')
    in_path = in_path[:-1]
    in_path = '/'.join(in_path)
    return(in_path+'/')


def read_table(file, sep='\t'):
    return(make_table(read_file(file,'lines'),sep))
    
def write_table(table, out_file, sep = '\t'):
    make_file(flatten_2D_table(table,sep), out_file)
    

def import_dict(f):
    f=open(f,'rb')
    d=pickle.load(f)
    f.close()
    return(d)

def save_dict(d,path):
    f=open(path,'wb')
    pickle.dump(d,f)
    f.close()

def cmd(in_message, com=True):
    print('\n',in_message)
    time.sleep(.25)
    #return()
    if com:
        Popen(in_message,shell=True).communicate()
    else:
        Popen(in_message,shell=True)



##############################################################



###############################################################
###############################################################

def main(args):
    ##################################################################
    ##################################################################
    ###################### clustering ################################
    ##################################################################
    ##################################################################
    args.infile = os.path.realpath(args.infile)

    temp_dir = get_file_path(args.infile)
    all_calls = [' '.join(sys.argv)]
    ##################################################################
    def do_the_clustering(temp_dir, args, out_adj_pos, all_calls, cor_stats_file=None, neg_cor_cutoff = 1):
        #global args, out_adj_pos, all_calls
        print('performing clustering')
        ## if we're doing the clustering build up the string for the command line

        if args.do_dispersion:
            if not os.path.isdir(temp_dir+'/sample_clustering_and_summary'):
                cmd('mkdir '+temp_dir+'/sample_clustering_and_summary')
            dispersion_str = 'pyminer_get_dispersion.py -i '+args.infile
            dispersion_str += ' -out_dir '+temp_dir+'/sample_clustering_and_summary/dispersion/'
            if args.h5py:
                dispersion_str += ' -hdf5 '
                dispersion_str += ' -ID_list '+args.ID_list
            all_calls.append(dispersion_str)
            cmd(dispersion_str)

        cluster_string = 'clustering.py -i '+args.infile+' -species '+args.species
        if args.sc_clust:
            cluster_string += ' -sc_clust '
            cluster_string += ' -neg_cor_count_clust '+str(temp+'/sum_neg_cor.txt')
        if args.pre_determined_sample_k!=None:
            cluster_string += ' -sample_k_known '+str(args.pre_determined_sample_k)
        cluster_string += ' -sample_cluster_iter '+str(args.sample_cluster_iter)
        if args.var_norm_for_clust:
            cluster_string += ' -var_norm '
        if args.neg_cor_clust and not args.sc_clust:
            ### this is broken...
            ## need to figure this out...
            cluster_string += ' -neg_cor_clust '+str(cor_stats_file)
            cluster_string += ' -neg_cor_cutoff '+str(neg_cor_cutoff)
            cluster_string += ' -first_neg_neighbor '
            cluster_string += ' -pos_adj_list '+out_adj_pos
        if not args.spearman_clust:
            cluster_string += ' -no_spearman_clust '
            
        if args.ap_clust:
            cluster_string += ' -ap '
        if args.louvain_clust:
            cluster_string += ' -louvain_clust '
        if args.h5py:
            cluster_string += ' -hdf5 '
            if args.ID_list == None:
                sys.exit('for hdf5, we need the ID_list argument')
            else:
                cluster_string += ' -ID_list '+args.ID_list
            if args.columns == None:
                sys.exit('for hdf5, we need the ID_list argument')
            else:
                cluster_string += ' -columns '+args.columns

        if args.merge:
            cluster_string += ' -merge '

        if args.do_dispersion:
            cluster_string += " -clust_on_genes "+temp_dir+'/sample_clustering_and_summary/dispersion/locally_overdispersed_boolean_table.txt'

        if args.cluster_on_genes!=None:
            cluster_string += " -clust_on_genes "+args.cluster_on_genes

        if args.leave_mito_ribo:
            cluster_string += " -leave_mito_ribo "

        cluster_string += ' -rand_seed '+str(args.rand_seed)
        if args.manual_sample_groups!=None:
            cluster_string += ' -manual_sample_groups '+args.manual_sample_groups
        cluster_string += ' -block_size '+str(args.block_size)
        all_calls.append(cluster_string)
        cmd(cluster_string)


    ##################################################################


    if args.h5py:
        do_h5_py=True
        import h5py
        if args.ID_list==None:
        	sys.exit("\nwhen using the hdf5 format, you must\nsupply the -ID_list as another argument")
        title = ''
    else:
        do_h5_py=False


    temp=str(args.infile).split('/')
    temp=('/').join(temp[:-1])

    if args.ID_list!=None:
        ID_list_path=args.ID_list
        IDlist = read_file(ID_list_path,'lines')
        make_file('\n'.join(IDlist),temp+'/ID_list.txt')
    else:
        IDlist=[]
        first = True
        for line in fileinput.input(args.infile):
            if first:
                first = False
                pass
            else:
                temp_line = line.split('\t')
                IDlist.append(temp_line[0])
        make_file('\n'.join(IDlist),temp+'/ID_list.txt')


    ID_hash={gene:i for i, gene in enumerate(IDlist)}
    id_file_name = temp+'/ID_list.txt'
    ###########################################################################
    ###########################################################################
    ################### call all of the other scripts #########################
    ###########################################################################
    ###########################################################################

    ################################################
    ## get gene annotations
    annotation_prefix = temp+'/annotations'
    annotation_call = 'pyminer_gprofiler_converter.py -i '+id_file_name
    annotation_call += ' -o '+annotation_prefix
    annotation_call += ' -s '+args.species
    annotation_call += ' -annotations'
    cmd(annotation_call)

    ## get the human orthologues if we don't have human ids
    if args.species != 'hsapiens':
        orthologue_prefix = temp+'/human_orthologues'
        convert_call = 'pyminer_gprofiler_converter.py -i '+id_file_name
        convert_call += ' -o '+orthologue_prefix
        convert_call += ' -s '+args.species
        cmd(convert_call)

    ################################################
    ## 3: build the network(s)
    ## if we're doing a single network:
    if args.rho_cutoff!=None:
        if args.rho_cutoff > 1.0 or args.rho_cutoff<0:
            sys.exit('please use a rho cutoff between zero and one. We recommend not using this argument at all to allow for the built in False Positive Rate (FPR) algorithm to take over. If using scRNAseq, try something in the range of 0.25-0.35. For bulk RNAseq, try something closer to 0.70 - 0.90.')

    out_adj = temp+'/adj_list_rho.tsv'
    adj_call_string = 'mat_to_adj_list.py -i '+args.infile+' -o '+out_adj+' -id_list '+temp+'/ID_list.txt '
    if args.rho_cutoff!=None:
        adj_call_string+=' -rho '+str(args.rho_cutoff)
    if args.h5py:
        adj_call_string += ' -hdf5 '
    adj_call_string += ' -block_size '+str(args.block_size)
    if args.sc_clust:
        adj_call_string += ' -sc_clust '
    if args.prop:
        adj_call_string += ' -prop '
    adj_call_string += " -FPR_multiple "+str(args.FPR_multiple)
    adj_call_string += ' -rand_seed '+str(args.rand_seed)
    all_calls.append(adj_call_string)
    cmd(adj_call_string)

    #######################
    ## 3.5: subset the positive/negative correlations

    pos_neg_count = np.zeros((len(IDlist),2))

    out_adj_pos = out_adj[:-4]+'_dedup_pos.tsv'
    cmd('rm '+out_adj_pos)
    pos_f_out=open(out_adj_pos,'a')

    out_adj_neg = out_adj[:-4]+'_dedup_neg.tsv'
    cmd('rm '+out_adj_neg)
    neg_f_out=open(out_adj_neg,'a')

    counter = 0
    first = True
    for line in fileinput.input(out_adj[:-4]+'_dedup.tsv'):
        if first:
            print('reading ',out_adj[:-4]+'_dedup.tsv')
            #print(float(temp_line[-1]))
            pos_f_out.write(line)
            neg_f_out.write(line)
            first = False
        else:
            temp_line = strip_split(line)
            #print(float(temp_line[-1]))
            try:
                float(temp_line[-1])
            except:
                print('error at line:',counter)
                print('\t',line)
            if float(temp_line[-1])>0:
                pos_f_out.write(line)
                pos_neg_count[ID_hash[temp_line[0]]][0]+=1
                pos_neg_count[ID_hash[temp_line[1]]][0]+=1
            else:
                neg_f_out.write(line)
                pos_neg_count[ID_hash[temp_line[0]]][1]+=1
                pos_neg_count[ID_hash[temp_line[1]]][1]+=1
        counter+=1
    pos_f_out.close()
    neg_f_out.close()

    ## log some stats about how many positive and negative relationships there are
    pos_neg_count = pos_neg_count.tolist()
    for i in range(0,len(pos_neg_count)):
        pos_neg_count[i] = [IDlist[i]] + pos_neg_count[i]

    pos_neg_count = [['ID','num_pos_cor','num_neg_cor']]+ pos_neg_count

    write_table(pos_neg_count, temp+'/positive_negative_cor_counts.txt')

    ################################################
    ## do feature selection



    ########################## clustering #############################
    if args.manual_sample_groups!=None:
        do_the_clustering(temp_dir, args, out_adj_pos, all_calls)
        args.manual_sample_groups = temp_dir+'sample_clustering_and_summary/sample_k_means_groups.txt'
    elif args.neg_cor_clust:
        do_the_clustering(temp_dir, args, out_adj_pos, all_calls,cor_stats_file=temp+'/positive_negative_cor_counts.txt', neg_cor_cutoff = args.neg_cor_cutoff)
        args.manual_sample_groups = temp_dir+'sample_clustering_and_summary/sample_k_means_groups.txt'
    else:
        ## do the clustering if it is needed
        if not args.h5py or not args.neg_cor_clust:
            if args.do_sample_clustering:
                do_the_clustering(temp_dir, args, out_adj_pos, all_calls)
                args.manual_sample_groups = temp_dir+'sample_clustering_and_summary/sample_k_means_groups.txt'
            elif args.manual_sample_groups==None:
                ## if we aren't going to be doing the clustering, we'll need to pretend that these are all from the same group
                ## to do this we'll write the sample_k_means_groups_file 
                if not args.h5py:
                    first = True
                    for temp_line in fileinput.input(args.infile):
                        if first:
                            title = np.array((temp_line.strip('\n')).split('\t'), dtype='U32')
                            first = False
                            break
                    fileinput.close()
                else:
                    title = np.array(read_file(args.columns,'lines'))
                ##############################################
                sample_names = title[1:]
                sample_k_table = [0]*(len(sample_names))
                sample_k_table = list(zip(sample_names, [0]*len(sample_names)))

                ##make the manual sample groups true, and make the file
                out_sample_group_table = np.array(sample_k_table)
                cmd('mkdir '+temp_dir+'sample_clustering_and_summary')
                write_table(out_sample_group_table,temp_dir+'sample_clustering_and_summary/sample_k_means_groups.txt')

                ## after we finish the clustering (or not), re-direct the manual sample groups to the 
                ## output sample group identity file
                args.manual_sample_groups = temp_dir+'sample_clustering_and_summary/sample_k_means_groups.txt'




        if args.manual_sample_groups!=None:
            manual_sample_groups_file = args.manual_sample_groups
        #    manual_sample_groups = read_table(manual_sample_groups_file)
            sample_k_table = read_table(manual_sample_groups_file)
            manual_sample_groups = True
            sample_cluster_ids = []
            for i in range(0,len(sample_k_table)):
                
                ## THIS IS IMPORTANT
                ## here we assume that the samples are all listed in the same order as in '-infile'
                ## we also assume that the group indexing starts at 0
                sample_cluster_ids.append(sample_k_table[i][1])
            sample_cluster_ids = list(map(int,sample_cluster_ids))
            sample_k_lists = []
            for i in range(0,max(sample_cluster_ids)+1):
                sample_k_lists.append([])
            #print(len(sample_k_lists))
            ## now populate the list of lists
            for i in range(0,len(sample_cluster_ids)):
                ## this appends the sample index to 
                #print(sample_cluster_ids[i])
                sample_k_lists[sample_cluster_ids[i]].append(i)
        else:
            args.manual_sample_groups = False

        import random
        random.seed(args.rand_seed)
        np.random.seed(args.rand_seed)

        many_variables = False
        very_big_file = False


        if '-big_file_do_z' in sys.argv:
            big_file_do_z=True
        else:
            big_file_do_z=False

    if args.sample_cluster_only:
        sys.exit('done with clustering')

    ################################################
    ## : enrichment & significance
    if True:#not args.h5py:
        stats_call = 'get_stats.py -i '+args.infile+' -sample_groups '+args.manual_sample_groups+' -FDR '+str(args.FDR_cutoff)+' -Zscore '+str(args.zscore_cutoff)
        if args.species != None:
            stats_call += str(' -s '+args.species)
        if args.h5py:
            stats_call += ' -hdf5 '
            if args.ID_list == None:
                sys.exit('for hdf5, we need the ID_list argument')
            else:
                stats_call += ' -ID_list '+args.ID_list
            if args.columns == None:
                sys.exit('for hdf5, we need the ID_list argument')
            else:
                stats_call += ' -columns '+args.columns
        if args.processes is not None:
            stats_call += ' -processes '+str(args.processes)
        all_calls.append(stats_call)
        cmd(stats_call)
        if not os.path.isdir(temp+"/sample_clustering_and_summary/significance/"):
            cmd("mkdir "+temp+"/sample_clustering_and_summary/significance/")
        make_file(stats_call,temp+"/sample_clustering_and_summary/significance/stats_call.txt")

    ################################################
    ## combine the gprofiler results
    if os.path.isdir(temp+"/sample_clustering_and_summary/significance/gprofiler"):
        combine_gprofiler_str = "combine_gprofiler_results.py -i "+temp+"/sample_clustering_and_summary/significance/gprofiler/"
        combine_gprofiler_str +=" -o "+temp+"/sample_clustering_and_summary/significance/"
        all_calls.append(combine_gprofiler_str)
        cmd(combine_gprofiler_str)



    ################################################
    ## get and plot highly expressed marker genes
    group_means_file = temp+"/sample_clustering_and_summary/"+"k_group_means.txt"
    aov_file = temp+"/sample_clustering_and_summary/significance/"+"groups_1way_anova_results.txt"
    annotation_dict = temp+"/annotations.pkl"
    high_marker_out_dir = temp+"/sample_clustering_and_summary/significance/high_markers/"
    cluster_pkl = temp+"/sample_clustering_and_summary/clustering_plots.pkl"

    if os.path.isdir(temp+"/sample_clustering_and_summary/significance/"):
        cmd("mkdir "+high_marker_out_dir)
        get_markers_call = 'get_high_marker_genes.py -m '+group_means_file
        get_markers_call += ' -sig '+aov_file
        get_markers_call += ' -out '+high_marker_out_dir
        get_markers_call += ' -annotation_dict '+annotation_dict
        cmd(get_markers_call)
        all_calls.append(get_markers_call)

        plot_subset_file = high_marker_out_dir+"subset_input.txt"
        marker_plot_call = "plot_gene_subset.py -i "+args.infile+" -noi "+plot_subset_file
        marker_plot_call += " -heatmap "
        marker_plot_call += " -ids "+id_file_name
        marker_plot_call += " -o "+high_marker_out_dir
        marker_plot_call += " -cluster_pkl "+cluster_pkl
        if args.h5py:
            marker_plot_call += ' -hdf5 '
            if args.ID_list == None:
                sys.exit('for hdf5, we need the ID_list argument')
            else:
                marker_plot_call += ' -ID_list '+args.ID_list
            if args.columns == None:
                sys.exit('for hdf5, we need the ID_list argument')
            else:
                marker_plot_call += ' -columns '+args.columns
        cmd(marker_plot_call)
        all_calls.append(marker_plot_call)

    ################################################
    ## 4: network analysis
    ## check if the enrichment file exists
    #cmd('mkdir '+temp+'/graphs')
    cmd('mkdir '+temp+'/pos_cor_graphs')

    full_network_plot_call = "plot_network.py -id_list "+temp+'/ID_list.txt'+" -adj "+out_adj+" -plot_all  -out_dir "+temp+"/graphs"
    pos_cor_network_call = "plot_network.py -id_list "+temp+'/ID_list.txt'+" -adj "+out_adj_pos+" -plot_all  -out_dir "+temp+"/pos_cor_graphs"

    if os.path.isfile(temp+"/sample_clustering_and_summary/k_group_enrichment.txt"):
        full_network_plot_call+=" -node_attrs "+temp+"/sample_clustering_and_summary/k_group_enrichment.txt -range "+str(args.zscore_cutoff)+",-"+str(args.zscore_cutoff)
        pos_cor_network_call+=" -node_attrs "+temp+"/sample_clustering_and_summary/k_group_enrichment.txt -range "+str(args.zscore_cutoff)+",-"+str(args.zscore_cutoff)

    #cmd(full_network_plot_call)
    all_calls.append(pos_cor_network_call)
    cmd(pos_cor_network_call)



    ################################################
    ## analyze genes of interest

    def process_noi(nodes_of_interest):
        noi_table=[]
        for i in range(0,len(nodes_of_interest)):
            if '\t' in nodes_of_interest[i]:
                noi_table.append(nodes_of_interest[i].split('\t'))
            elif '::' in nodes_of_interest[i]:
                noi_table.append(nodes_of_interest[i].split('::'))
            else:
                noi_table.append([nodes_of_interest[i],nodes_of_interest[i]])

        # print(noi_table)
        ## also make the ailias look-up dictionary
        alias_dict={}
        ## make the string that would be passed through to command line if needed
        pass_through_arg =""
        ## and the final list of nodes in the dataset
        final_nodes = []
        for i in range(0,len(noi_table)):
            # print(i)
            # print(noi_table[i])
            alias_dict[noi_table[i][0]]=noi_table[i][1]
            pass_through_arg+=","+noi_table[i][0]+'::'+noi_table[i][1]
            final_nodes.append(noi_table[i][0])
        ## remove the leading comma
        pass_through_arg=pass_through_arg[1:]

        return(alias_dict, pass_through_arg, final_nodes)



    if args.noi != None:
        ## first check if it's a path
        if os.path.isfile(args.noi):
            noi_list = read_file(args.noi,'lines')
        else:
            ## check if it's a comma separated list of ids
            temp_nodes_of_interest = args.noi.split(',')

        alias_dict, pass_through_arg, temp_nodes_of_interest = process_noi(noi_list)
        
        print(temp_nodes_of_interest)
        

        final_noi_list = []
        for noi in temp_nodes_of_interest:
            if noi in IDlist:
                final_noi_list.append(str(noi))
            else:
                print("couldn't find "+str(noi)+" in the IDs...")
        print('getting degrees of separation for:')
        for noi in final_noi_list:
            print('\t',noi)

        #print(final_noi_list)
        

        ## now run the call
        noi_call = "get_degrees_of_separation_from_adjacency_list.py -adj "+out_adj_pos
        noi_call += " -ids "+temp+"/ID_list.txt"
        noi_call += " -noi "+','.join(final_noi_list)
        noi_call += " -out_dir "+temp+'/genes_of_interest'
        cmd('mkdir '+temp+'/genes_of_interest')
        all_calls.append(noi_call)
        cmd(noi_call)
        

        ## if the plots file is there plot the color over the positions
        plot_subset_call = 'plot_gene_subset.py -i '+args.infile
        plot_subset_call += " -out_dir "+temp+'/genes_of_interest'
        plot_subset_call += " -noi "+pass_through_arg
        plot_subset_call += " -id_list "+temp+'/ID_list.txt'
        plot_subset_call += " -cluster_pkl "+temp+"/sample_clustering_and_summary/clustering_plots.pkl"
        if args.h5py:
            plot_subset_call += ' -hdf5 '
            plot_subset_call += ' -cols '+temp+'/column_IDs.txt'
        all_calls.append(plot_subset_call)
        cmd(plot_subset_call)





    #################################################
    if not args.no_cell_signals:
        cmd('mkdir '+temp+'/autocrine_paracrine_signaling/')
        cell_signals_call = "cell_signals.py -i "+temp+'/sample_clustering_and_summary/significance/significant_and_enriched_boolean_table.txt '
        cell_signals_call += ' -o '+temp+'/autocrine_paracrine_signaling/'
        cell_signals_call += ' -species '+args.species
        if args.string_db_dir[-1]!='/':
            args.string_db_dir+='/'
        cell_signals_call += ' -sdb '+args.string_db_dir
        all_calls.append(cell_signals_call)
        cmd(cell_signals_call)
        if os.path.isfile(temp+'/autocrine_paracrine_signaling/all_cell_type_specific_interactions_gprofiler.txt'):
            combine_gprofiler2_str="combine_gprofiler_results.py -i "+temp+'/autocrine_paracrine_signaling/all_cell_type_specific_interactions_gprofiler.txt'
            combine_gprofiler2_str+=' -o '+temp+'/autocrine_paracrine_signaling/'
            all_calls.append(combine_gprofiler2_str)
            cmd(combine_gprofiler2_str)


    #################################################
    if os.path.isfile(temp+'/pos_cor_graphs/communities.txt'):
        community_analysis = "pyminer_analyze_communities.py -community_table "+temp+"/pos_cor_graphs/communities.txt -out_dir "+temp+"/pos_cor_graphs/community_analysis -z_score_table "+temp+"/sample_clustering_and_summary/k_group_enrichment.txt -species "+args.species
        community_analysis += ' -ad '+temp+'/annotations.pkl'
        cmd(community_analysis)


    ################################################
    ## summarize the results via the website generator
    website_call = "make_website.py -i "+temp_dir
    all_calls.append(website_call)


    make_file('\n'.join(all_calls),temp+'/PyMINEr_call_log.txt')


    cmd(website_call, com=False)



if __name__ == '__main__':
    parser = argparse.ArgumentParser()


    ## global arguments
    parser.add_argument(
        '-infile','-in','-i','-input',
        dest='infile',
        type=str)

    parser.add_argument(
        "-hdf5",'-h5py',
        help='if the "-infile" is an hdf5 file whose main array name is "infile". This argument also requires "-ID_list" as a separate argument.',
        dest='h5py',
        action='store_true',
        default = False)
    parser.add_argument(
        '-ID_list','-IDs','-ids',
        help = "A new line delimited file with no title line that contains the IDs associated with each row, in the correct order",
        dest='ID_list',
        type=str)
    parser.add_argument(
        '-columns','-cols',
        help = "A new line delimited file with no title line that contains the sample IDs associated with each column, in the correct order",
        dest='columns',
        type=str)

    parser.add_argument("-rand_seed",
        type = int,
        default = 12345)


    ########################################################
    ## define which steps to do (ie: sample clustering, build networks, etc)
    parser.add_argument(
        "-networks_only",'-net_only',
        help='only find the networks and write adjacency list(s)',
        dest='mineOnly',
        action='store_true',
        default = False)

    parser.add_argument(
        "-sample_cluster_only",
        help='only do PyMINEr clustering',
        dest='sample_cluster_only',
        action='store_true',
        default = False)

    parser.add_argument(
        "-no_sample_cluster","-no_sample_clust",
        help='skip the sample clustering',
        dest='do_sample_clustering',
        action='store_false',
        default = True)

    parser.add_argument(
        "-cluster_on_all",
        help='cluster on all variables, without looking at dispersion',
        dest='do_dispersion',
        action='store_false',
        default = True)


    parser.add_argument(
        "-cluster_on_genes","-clust_on_genes","-clust_genes",
        help='If there is a specific set of genes you want to cluster on, supply the gene ids in this text file.'
        )

    parser.add_argument(
        "-leave_mito_ribo",
        help='If we should leave the mitochondrial and ribosomal genes for clustering',
        action='store_true',
        default = False)



    # parser.add_argument(
    #     "-neg_cor_cutoff",
    #     help='for negative correlation based clustering, what should the cutoff be? (default = 20)',
    #     type = int,
    #     default = 15)

    parser.add_argument(
        "-beta_test",
        help='set the paramaters for beta_testing',
        dest = "sc_clust",
        action='store_true',
        default = False)


    ########################################################
    ## sample clustering options


    parser.add_argument(
        "-sample_k_clusters_known",'-sample_k_known',
        help='if you know how many groups there should be',
        dest = 'pre_determined_sample_k',
        type = int)

    parser.add_argument(
        "-sample_cluster_iter","-clust_iter",
        help='How many iterations of clustering should we do. This can take some time, but higher iterations will give better results. Default = 10',
        dest = 'sample_cluster_iter',
        type = int,
        default = 10)
    parser.add_argument(
        "-var_norm_for_clust","-var_norm",
        help='normalize the variables for sample clustering',
        dest='var_norm_for_clust',
        action='store_true',
        default = True)
    parser.add_argument(
        "-no_var_norm",
        help='do not normalize the variables for sample clustering',
        dest='var_norm_for_clust',
        action='store_false',
        default = True)
    parser.add_argument(
        "-neg_cor_clust",
        help='cluster based on variables with negative correlations',
        action='store_true',
        default = False)
    parser.add_argument(
        "-no_spearman_clust",
        help='do not use the spearman similarity matrix to do the clustering. The default is to use the symetric Spearman correlation matrix of all samples against each other.',
        dest='spearman_clust',
        action='store_false',
        default = True)
    parser.add_argument(
        "-ap_clust", '-ap',
        help='if you want to use affinity propagation clustering',
        action='store_true',
        default = False)
    parser.add_argument(
        "-louvain_clust",
        help='if you want to use louvain clustering',
        action='store_true',
        default = False)
    parser.add_argument(
        "-merge", 
        help='if you want to perform cluster merger analysis',
        action='store_true',
        default = False)


    parser.add_argument(
        "-manual_sample_groups",
        help='if you know how which samples belong to which groups, feed in a file that has the samples in the first column, and their group number (index starting at 0), in the second column. The IDs must be in the same order as in the infile too.',
        dest = 'manual_sample_groups',
        type = str)

    parser.add_argument(
        "-species",
        help='species to use for gProfiler analysis',
        dest = 'species',
        default = 'hsapiens',
        type = str)

    ########################################################
    ## enrichment options
    parser.add_argument(
        "-anova_FDR_cutoff","-FDR_cutoff",
        help='The Benjamini Hochberg corrected FDR value cutoff for calling a variable significantly different between groups.',
        dest = 'FDR_cutoff',
        type = float,
        default = 0.05)

    parser.add_argument(
        "-zscore_cutoff",
        help='The Z-score cutoff for calling a variable enriched in a given group.',
        dest = 'zscore_cutoff',
        type = float,
        default = 2.0)

    parser.add_argument(
        "-processes",
        help='The number of processes for running the stats',
        dest = 'processes',
        type = int,
        default = None)

    ########################################################
    ## network detection options
    parser.add_argument("-spearman_only",
        dest = "spearman_only",
        action = 'store_true',
        default = False)

    parser.add_argument("-rho_cutoff",'-rho',
        dest = "rho_cutoff",
        help = "Suggested to leave this blank, but you can provide your own cutoff instead of allowing PyMINEr to set it for you. This is the absolute value of the spearman rho to use for a cutoff. If using scRNAseq, try something in the range of 0.25-0.35. For bulk RNAseq, try something closer to 0.70 - 0.90.",
        type = float)


    parser.add_argument("-prop","-proportaionality",
        action = 'store_true',
        default = False, 
        help="if you only want to use proportionality instead of Spearman")


    parser.add_argument("-block_size",
        help = 'how many variables will be used at once for correlation analyzis; this helps keep memory requirements down and prevents a segmentation fault memory error',
         type = int,
         default = 5000)

    parser.add_argument(
        "-FPR_multiple",'-mult',
        help="the multiple to use for FPR based anti-correlation feature selection. Only use this if you really know what you're doing.",
        type = int,
        default = 15)


    ########################################################
    ## if you already have a(n) adjacency list(s)
    # parser.add_argument(
    #   '-network1','-adj1','-adj_list1','-network','-net1','-in_mine1',
    #   help = "if you already have an adjacency list, we can use that instead of building up a network from scratch.",
    #   dest='in_mine1',
    #   type=str)
    # parser.add_argument(
    #   '-network2','-adj2','-adj_list2','-net2','-in_mine2',
    #   help = "if you have a second adjacency list for a network comparison, pass the path to this option.",
    #   dest='in_mine2',
    #   type=str)
    # parser.add_argument(
    #   '-name1',
    #   dest='control_name',
    #   default = 'network1',
    #   type=str)
    # parser.add_argument(
    #   '-name2',
    #   dest='disease',
    #   default = 'network2',
    #   type=str)

    ###################################################
    ## for getting degree of separation from genes of iterest
    parser.add_argument(
        "-genes_of_interest",'-nodes_of_interest','-goi','-noi',
        help="if you have guesses about genes of interest a priori, we'll get the degree of separation from those genes in the primary analysis. This can either be a file with genes each on a new line, or it can be a comma separated list of genes.",
        dest = "noi",
        type = str)

    ###################################################
    ## cell signals options
    parser.add_argument(
        "-no_cell_signals",
        help="if you don't want to do the autocrine paracrine signaling network prediction",
        action = 'store_true',
        default = False)
    parser.add_argument(
        "-string_db_dir",'-stringdb_dir','-sdb',
        help="If the stringDB directory isn't in /usr/local/lib/cell_signals/, you'll have to tell PyMINEr where to look. This is often the case if you don't have a global (sudo) install of PyMINEr.",
        dest = "string_db_dir",
        default = '/usr/local/lib/cell_signals/',
        type = str)

    ############################################
    ## aesthetics options
    parser.add_argument(
        "-dpi",
        help='dots per inch for the output figures',
        dest = "dpi",
        type = int,
        default = 360)

    parser.add_argument(
        "-verbose",
        help='prints out some extra lines; this is primarily for troubleshooting',
        dest='verbose',
        action='store_true',
        default = False)
    args = parser.parse_args()

    if args.cluster_on_genes:
        #args.leave_mito_ribo = True
        args.do_dispersion = False

    if args.neg_cor_clust or args.sc_clust:
        args.do_dispersion = False

    #print(dir(args))
    ##################################################################
    # if args.dividing_col != 0:
    #     single_network=False
    #     mannwhit=True
    # else:
    #     single_network=True
    #     mannwhit=False

    ###################
    # log_fold=True
    # log_expression=False
    # rainbow=True
    # no_express=False
    # express_only=False
    # do_spearman = True
    # co_regulatory_network=True

    # ## page rank options that won't change
    # page_ranks=True
    # pr_iters=50 ## this is a totally arbitrary default
    # pr_alpha=0.85
    # weighted_analysis=True

    # ## clustering options that won't change
    # cluster_prob=False
    # first_prob = True

    # if args.manual_sample_groups!=None:
    #     args.do_sample_clustering=False
    main(args)



#!/usr/bin/env python3
##import dependency libraries
import sys,time,glob,os,pickle,fileinput,argparse
from subprocess import Popen
from operator import itemgetter
import gc, fileinput
import numpy as np
from scipy import special 
#import pandas as pd
##############################################################
## basic function library
def read_file(tempFile,linesOraw='lines',quiet=False):
    if not quiet:
        print('reading',tempFile)
    f=open(tempFile,'r')
    if linesOraw=='lines':
        lines=f.readlines()
        for i in range(0,len(lines)):
            lines[i]=lines[i].strip('\n')
    elif linesOraw=='raw':
        lines=f.read()
    f.close()
    return(lines)

def make_file(contents,path):
    f=open(path,'w')
    if isinstance(contents,list):
        f.writelines(contents)
    elif isinstance(contents,str):
        f.write(contents)
    f.close()

    
def flatten_2D_table(table,delim):
    #print(type(table))
    if str(type(table))=="<class 'numpy.ndarray'>":
        out=[]
        for i in range(0,len(table)):
            out.append([])
            for j in range(0,len(table[i])):
                try:
                    str(table[i][j])
                except:
                    print(table[i][j])
                else:
                    out[i].append(str(table[i][j]))
            out[i]=delim.join(out[i])+'\n'
        return(out)
    else:
        for i in range(0,len(table)):
            for j in range(0,len(table[i])):
                try:
                    str(table[i][j])
                except:
                    print(table[i][j])
                else:
                    table[i][j]=str(table[i][j])
            table[i]=delim.join(table[i])+'\n'
    #print(table[0])
        return(table)

def strip_split(line, delim = '\t'):
    return(line.strip('\n').split(delim))

def make_table(lines,delim):
    for i in range(0,len(lines)):
        lines[i]=lines[i].strip()
        lines[i]=lines[i].split(delim)
        for j in range(0,len(lines[i])):
            try:
                float(lines[i][j])
            except:
                lines[i][j]=lines[i][j].replace('"','')
            else:
                lines[i][j]=float(lines[i][j])
    return(lines)


def get_file_path(in_path):
    in_path = in_path.split('/')
    in_path = in_path[:-1]
    in_path = '/'.join(in_path)
    return(in_path+'/')


def read_table(file, sep='\t'):
    return(make_table(read_file(file,'lines'),sep))
    
def write_table(table, out_file, sep = '\t'):
    make_file(flatten_2D_table(table,sep), out_file)
    

def import_dict(f):
    f=open(f,'rb')
    d=pickle.load(f)
    f.close()
    return(d)

def save_dict(d,path):
    f=open(path,'wb')
    pickle.dump(d,f)
    f.close()

def cmd(in_message, com=True):
    print(in_message)
    time.sleep(.25)
    if com:
        Popen(in_message,shell=True).communicate()
    else:
        Popen(in_message,shell=True)



##############################################################



###############################################################
###############################################################

parser = argparse.ArgumentParser()


## global arguments
parser.add_argument(
	'-in_dir','-in','-i','-input',
	dest='in_dir',
	type=str)
parser.add_argument(
	'-out_dir','-o','-out',
	dest='out',
	type=str)

args = parser.parse_args()
###############################################################
###############################################################
## functions for calculating the final ranking metric
def norm(vect):
	vect = vect - min(vect)
	vect = vect / max(vect)
	return(vect)

def kl_from_gaus(vect,fudge=1e-4,bins=5):
    ## first make the default Gaussian distribution probability vector
    vect = np.array(vect)
    # gaus_bins = np.histogram(np.random.normal(size=max([10000,len(vect)])),bins=len(vect),density=True)[0]
    # observed = np.histogram(norm(vect), bins = len(vect))[0]
    gaus_bins = np.histogram(np.random.normal(size=max([10000,len(vect)])),bins=bins,density=True)[0]
    observed = np.histogram(norm(vect), bins = bins)[0]

    ## do the kl divergence, but add the fudge factor to prevent inf
    full_kl = special.kl_div(gaus_bins+fudge,observed+fudge)
    return(np.sum(full_kl))


def get_rank_metric(vect):
	sum_kl = kl_from_gaus(vect)
	# multiply by the range
	rank_metric = (max(vect)-min(vect)) * sum_kl
	return(rank_metric)



###############################################################
###############################################################
## if the gprofiler results were generated by cell_signals it will be in a different format than from pyminer_gprofile.py
## the first thing we have to do is figure out which one it is.
## PyMINEr will be a directory, cell signals will be a single file.

if os.path.isdir(args.in_dir):
    print(args.in_dir,'is a directory')
    pyminer = True
    cell_signals = False
elif os.path.isfile(args.in_dir):
    print(args.in_dir,'is a file')
    cell_signals = True
    pyminer = False
else:
    sys.exit("Huston, we have a problem. The file or directory that was passed through couldn't be found/read")


###############################################################
def get_unique_terms(term_ids):
        ## make the terms unique
    print('all terms',len(term_ids))
    term_ids = sorted(list(set(term_ids)))
    print('unqiue',len(term_ids))

    ## catelogue the indices of the unique terms
    term_index_hash = {j:i for i, j in enumerate(term_ids)}

    return(term_ids, term_index_hash)

def get_empty_annotations(term_ids):
    ## make the annotation leader columns
    annotations = []
    for i in range(0,len(term_ids)):
        annotations.append([term_ids[i],None, None])
    return(annotations)

###############################################################
if cell_signals:
    ## catelogue the terms and groups
    all_groups = []
    term_ids = []
    first = True
    for line in fileinput.input(args.in_dir):
        if first:
            first = False
        else:
            ## get the list of term ids and group ids
            temp_line = strip_split(line)
            all_groups.append(temp_line[0]+"_&_"+temp_line[1])
            term_ids.append(temp_line[10])
    fileinput.close()

    all_groups = sorted(list(set(all_groups)))
    all_groups_index_hash = {j:i for i, j in enumerate(all_groups)}
    print(all_groups)
    term_ids, term_index_hash = get_unique_terms(term_ids)

    ## if there aren't any results, there's really nothing left to do...
    if len(term_ids)==0:
        sys.exit('no gprofiler results...')

    ## get the annotations
    annotations = get_empty_annotations(term_ids)

    ## populate the annotations
    first = True
    for line in fileinput.input(args.in_dir):
        if first:
            first = False
        else:
            ## get the list of term ids and group ids
            temp_line = strip_split(line)
            term_row = term_index_hash[temp_line[10]]
            annotations[term_row][0]=temp_line[10]
            annotations[term_row][1]=temp_line[11]
            annotations[term_row][2]=temp_line[13]
    fileinput.close()

    for i in range(0,5):
        print(annotations[i])

    ## make the negative log10 p-value matrix
    neg_log_p = np.zeros((len(term_ids),len(all_groups)))
    first = True
    for line in fileinput.input(args.in_dir):
        if first:
            first = False
        else:
            ## get the list of term ids and group ids
            temp_line = strip_split(line)
            temp_group = temp_line[0]+"_&_"+temp_line[1]
            temp_col_idx = all_groups_index_hash[temp_group]
            temp_row_idx = term_index_hash[temp_line[10]]
            temp_neg_log_p = -1*np.log10(float(temp_line[4]))
            neg_log_p[temp_row_idx,temp_col_idx] = temp_neg_log_p
    fileinput.close()     

###############################################################
###############################################################
if pyminer:
    ## go through all of the files logging all of the pathways that appear

    """
    p-value: 2
    term id: 8
    term name: 11
    """

    os.chdir(args.in_dir)
    all_groups = []
    for f in glob.glob("*gprofiler.txt"):
    	#print(f)
    	all_groups.append(f)

    all_groups = sorted(all_groups)

    ###################################################################
    ## go through the files & grab the unique term ids
    term_ids = []
    for f in all_groups:
    	first = True
    	for line in fileinput.input(f):
    		if first:
    			first = False
    		else:
    			temp_line = strip_split(line)
    			#print(temp_line[2],temp_line[8],temp_line[11])
    			term_ids.append(temp_line[8])
    	fileinput.close()

    ## if there aren't any results, there's really nothing left to do...
    if len(term_ids)==0:
    	sys.exit('no gprofiler results...')


    term_ids, term_index_hash = get_unique_terms(term_ids)

    ###################################################################
    annotations = get_empty_annotations(term_ids)

    for f in all_groups:
        first = True
        for line in fileinput.input(f):
            if first:
                first = False
            else:
                temp_line = strip_split(line)
                term_idx = term_index_hash[temp_line[8]]
                annotations[term_idx][1] = temp_line[9]
                annotations[term_idx][2] = temp_line[11]

    for i in range(0,5):
    	print(annotations[i])


    ###################################################################

    ## make the neg log p-val matrix
    neg_log_p = np.zeros((len(term_ids),len(all_groups)))


    ## add the negative log10 p-values
    for f in range(0,len(all_groups)):
    	first = True
    	for line in fileinput.input(all_groups[f]):
    		if first:
    			first = False
    		else:
    			temp_line = strip_split(line)
    			term_idx = term_index_hash[temp_line[8]]
    			neg_log_p[term_idx, f] = -np.log10(float(temp_line[2]))


    print(neg_log_p)

## normalize the -log10(p-values) within each group
def lin_norm_rows(in_mat,min_range=0,max_range=1):
    in_mat = np.transpose(np.array(in_mat))
    in_mat = in_mat - np.min(in_mat, axis = 0)
    in_mat = in_mat / np.max(in_mat, axis = 0)
    in_mat[np.isnan(in_mat)]=0
    return(np.transpose(in_mat))

## normalize the neg_log_p
norm_neg_log_p = np.transpose(lin_norm_rows(np.transpose(neg_log_p)))


###################################################################
## prioritize them based on the sum kl-divergance*range
def get_full_kl(in_mat, fudge = 1e-4):
    global term_ids
    kl_metric=[]
    for i in range(0,len(term_ids)):
       kl_metric.append(get_rank_metric(in_mat[i,:]))
    ## normalize kl metric
    kl_metric = kl_metric - min(kl_metric) + fudge
    kl_metric = kl_metric / max(kl_metric)

    return(kl_metric)

kl_metric = get_full_kl(neg_log_p)

reordering = sorted(range(len(kl_metric)), key=lambda k: kl_metric[k])[::-1]

## do the same for the individual class weights
individual_class_weights = norm_neg_log_p * np.transpose(np.array([kl_metric]))
individual_class_weights = np.transpose(lin_norm_rows(np.transpose(individual_class_weights)))
kl_individual = get_full_kl(individual_class_weights)

individual_reordering = sorted(range(len(kl_individual)), key=lambda k: kl_individual[k])[::-1]


###################################################################


## make the final output list
out_table_head = [['importance_metric', 'term_id','source','term_name']+all_groups]
out_table = out_table_head[:]
out_table_individual_importance = out_table_head[:]

for i in range(0,len(reordering)):
    temp_idx = reordering[i]
    out_table.append([kl_metric[temp_idx]]+annotations[temp_idx]+neg_log_p[temp_idx,:].tolist())

for i in range(0,len(individual_reordering)):
    temp_idx = individual_reordering[i]
    out_table_individual_importance.append([kl_individual[temp_idx]]+annotations[temp_idx]+individual_class_weights[temp_idx,:].tolist())


## write the file
if args.out[-1]!='/':
    args.out+='/'
    if os.path.isdir(args.out):
        pass
    else:
        cmd("mkdir "+args.out)

write_table(out_table,args.out+"combined_neg_log10p_gprofiler.txt")

write_table(out_table_individual_importance,args.out+'individual_class_importance.txt')


#!/usr/local/env python3

from gprofiler import GProfiler
from clustering import get_density, plot_densities, remove_zeros



##########################################################################
## if we are doing a negative correlation based run, subset the input matrix
## neg_cor plotting
def get_density(in_vect):
    density = gaussian_kde(in_vect)
    xs = np.arange(0,max(in_vect),.1)
    density.covariance_factor = lambda : 2.5
    density._compute_covariance()
    return(xs,density(xs))

def plot_densities(neg_cor_density_x, neg_cor_density_y, pos_cor_density_x, pos_cor_density_y,out_file):
    global args
    x_min = min([np.min(neg_cor_density_x),np.min(pos_cor_density_x)])
    y_min = min([np.min(neg_cor_density_y),np.min(pos_cor_density_y)])
    x_min = max([np.max(neg_cor_density_x),np.max(pos_cor_density_x)])
    y_min = max([np.max(neg_cor_density_y),np.max(pos_cor_density_y)])

    plt.clf()
    plt.plot(neg_cor_density_x, neg_cor_density_y, 
        label = 'negative correlation degree density',
        color = 'blue',
        linewidth = 3)
    plt.plot(pos_cor_density_x, pos_cor_density_y, 
        label = 'positive correlation degree density',
        color = 'red',
        linewidth = 3)
    plt.legend()
    plt.savefig(out_file,
        dpi=args.dpi,
        bbox_inches='tight')

def remove_zeros(in_vect):
    in_vect = np.array(in_vect).tolist()
    out_vect = []
    for i in range(0,len(in_vect)):
        if in_vect[i]!=0:
            out_vect.append(in_vect[i])
    return(out_vect)

def get_first_neighbors(passing_ids):
    global args
    first_neighbors = []
    for line in fileinput.input(args.pos_adj_list):
        temp_line = strip_split(line)
        if temp_line[0] in passing_ids or temp_line[1] in passing_ids:
            if temp_line[0] not in first_neighbors:
                first_neighbors.append(temp_line[0])
            if temp_line[1] not in first_neighbors:
                first_neighbors.append(temp_line[1])
    fileinput.close()
    return(first_neighbors)
## /neg_cor plotting
##########################################################################



#######################################################
## SECTION 1

def get_neg_cor_clust(neg_cor_clust):
    print('getting the negatively correlated variables')
    ## first we get the variables that pass muster (ie: have enough negative correlations)
    cor_stat_table = read_table(args.neg_cor_clust)
    num_genes_total = len(cor_stat_table)-1
    percentile_cutoff = 0.975
    top_num = int(num_genes_total * percentile_cutoff)

    ## make the plots
    cor_stat_array = np.array(cor_stat_table[:])

    ## get the vectors for the correlation counts
    neg_cor_vect = np.array(cor_stat_array[1:,2], dtype = float)
    pos_cor_vect = np.array(cor_stat_array[1:,1], dtype = float)
    neg_cor_density_x, neg_cor_density_y = get_density(neg_cor_vect)
    pos_cor_density_x, pos_cor_density_y = get_density(pos_cor_vect)

    ## plot them
    plot_densities(neg_cor_density_x, neg_cor_density_y, pos_cor_density_x, pos_cor_density_y,args.out_dir+'negative_positive_correlation_degree_density.png')
    neg_cor_list = sorted(neg_cor_vect[:].tolist())
    print(neg_cor_list[top_num])
    args.neg_cor_cutoff = neg_cor_list[top_num]


    ## do the same but after removing the zeros
    ## remove the zeros from these vectors
    neg_cor_vect_noZero = remove_zeros(neg_cor_vect[:])
    pos_cor_vect_noZero = remove_zeros(pos_cor_vect[:])

    ## get their densities and plot them
    nonZero_neg_cor_density_x, nonZero_neg_cor_density_y = get_density(neg_cor_vect_noZero)
    nonZero_pos_cor_density_x, nonZero_pos_cor_density_y = get_density(pos_cor_vect_noZero)
    plot_densities(nonZero_neg_cor_density_x, nonZero_neg_cor_density_y, nonZero_pos_cor_density_x, nonZero_pos_cor_density_y,args.out_dir+'non_zero_negative_positive_correlation_degree_density.png')

    ## come up with the dynamically determined cutoff
    neg_cor_vect_noZero = sorted(neg_cor_vect_noZero)
    percentile_cutoff = 0.85
    top_num = int(len(neg_cor_vect_noZero)*percentile_cutoff)
    print(neg_cor_vect_noZero[top_num])
    args.neg_cor_cutoff = neg_cor_vect_noZero[top_num]

    ########################################################
    passing_ids = []
    for i in range(1,len(cor_stat_table)):
        # start at 1 because of the title line
        if cor_stat_table[i][2] >= args.neg_cor_cutoff:
            # format is: id, # pos cor, # neg cor
            passing_ids.append(cor_stat_table[i][0])

    if len(passing_ids) == 0:
        ## if there aren't any anti-correlated genes, 
        ## then you probably only have one one cell type/state
        print('we only found one cell type/state')
    return()



if args.neg_cor_count_clust!=None:
    neg_cor_clust_table = read_table(args.neg_cor_count_clust)
    passing_ids = []
    for i in range(0,len(neg_cor_clust_table)):
        #print(neg_cor_clust_table[i][-1])
        if neg_cor_clust_table[i][-1] == "True":
            passing_ids.append(neg_cor_clust_table[i][0])



if args.neg_cor_clust!=None or args.neg_cor_count_clust!=None:
    ## to prevent giving too much weight to a single gene's expression, we can get the 
    ## first neighbor genes of the co-regulatory network, and add them to the genes 
    ## which got into the first pass negative correlation subset
    if args.first_neg_neighbor and False:
        print('getting the first neighbors of the highly negatively correlated genes')
        first_neighbors = get_first_neighbors(passing_ids)
        passing_ids = sorted(list(set(passing_ids+first_neighbors)))
        #print(passing_ids)

    if not args.do_hdf5:
        temp_full_expression_str = [full_expression_str[0]]
        #print(temp_full_expression_str)
        print(len(passing_ids),'of the variables had enough negative correlations to make the cut')
        ## collect the IDs that 
        usable_indices = []
        for i in range(1,len(full_expression_str)):
            ## start at 1 because of the title line
            if full_expression_str[i][0] in passing_ids:
                usable_indices.append(i-1)#-1 because of the title line
                #temp_full_expression_str.append(full_expression_str[i])
        #full_expression_str = temp_full_expression_str[:]
        #del temp_full_expression_str
        print('found',len(usable_indices),'usable indicies')
        #write_table(full_expression_str[:],args.out_dir+'neg_cor_mat_used_for_clustering.txt')
    else:
        usable_indices = []
        new_row_names = []
        for i in range(0,len(row_names)):
            if row_names[i] in passing_ids:
                usable_indices.append(i)
                new_row_names.append(row_names[i])
        print(len(usable_indices),'of the variables had enough negative correlations to make the cut')
## /SECTION 1
##########################################################################################################





##########################################################################################################
## SECTION 2



#####################
def process_dict(in_file, ensg_idx):
    out_dict = {}
    for i in range(0,len(in_file)):
        out_dict[in_file[i][ensg_idx]]=True
    return(out_dict)

def quick_search(in_dict,key):
    try:
        in_dict[key]
    except:
        return(False)
    else:
        return(True)
##############################


if args.clust_on_genes != None:
    print('picking out the specified genes')
    ## this can either be a table with the genes in the left most column or a new-line delimited list of only the gene ids
    ## if the last column is "False" it means this gene will not be included
    clust_on_genes = read_table(args.clust_on_genes)
    clust_gene_ids = []
    clust_gene_idxs = []
    if clust_on_genes == []:
        pass
    else:
        if type(clust_on_genes[0])==list:
            for i in range(0,len(clust_on_genes)):
                #print(clust_on_genes[i])
                try:
                    id_hash[str(clust_on_genes[i][0])]
                except:
                    print("couldn't find",clust_on_genes[i][0])
                else:
                    if clust_on_genes[i][-1] != "False":
                        clust_gene_ids.append(clust_on_genes[i][0])
        else:
            clust_gene_ids = clust_on_genes
        clust_gene_idxs = [id_hash[str(x)] for x in clust_gene_ids]
        if len(clust_gene_idxs)==0:
            print("couldn't find any of the genes in your input -clust_on_genes")
    usable_indices = clust_gene_idxs[:]
    print('\n\nwe found',len(clust_gene_idxs), 'genes in -clust_on_genes input')
else:
    ## if we've already filtered for negative correlations or something like that
    if 'usable_indices' in globals():
        clust_gene_idxs = usable_indices[:]
        #print(usable_indices)
        clust_gene_ids = np.array(ID_list)[usable_indices]
    ## If there weren't any other filters prior to this step
    else:
        clust_gene_ids = ID_list[:]
        clust_gene_idxs = np.arange(len(ID_list)).tolist()

#####
## set up a quick-search dictionary
print(len(usable_indices))
print(len(clust_gene_ids))
usable_id_hash = {key:value for value, key in enumerate(clust_gene_ids)}
##
#####

## remove the mito and ribo genes

def rm_ribo_mito(species, infile_original_dir):
    print("finding all of the ribosomal and mitchondrial genes to remove")
    ## if we'll remove the mitochondrial and ribsomal related genes first
    mitochondrial = ["GO:0044429","GO:0006390","GO:0005739","GO:0005743",
                     "GO:0070125","GO:0070126","GO:0005759","GO:0032543",
                     "GO:0044455","GO:0005761"]
    ribosome = ["GO:0005840","GO:0003735","GO:0022626","GO:0044391","GO:0006614",
                "GO:0006613","GO:0045047","GO:0000184","GO:0043043","GO:0006413",
                "GO:0022613","GO:0043604","GO:0015934","GO:0006415","GO:0015935",
                "GO:0072599","GO:0071826","GO:0042254","GO:0042273","GO:0042274",
                "GO:0006364","GO:0022618","GO:0005730","GO:0005791","GO:0098554",
                "GO:0019843"]
    ribo_mito_go = mitochondrial + ribosome
    
    gp = GProfiler('PyMINEr_'+str(random.randint(0,int(1e6))), want_header = True)
    results = gp.gconvert(ribo_mito_go,organism="hsapiens", target="ENSG")
    #results = gp.gconvert(ribo_mito_go,organism=args.species, target="ENSG")
    ## first go to the annotation file, and read that in.
    if args.species == 'hsapiens':
        if os.path.isfile(infile_original_dir+'annotations.tsv'):
            annotations = read_table(infile_original_dir+'annotations.tsv')
            ensg_idx = 3
    else:
        if os.path.isfile(infile_original_dir+'human_orthologues.tsv'):
            annotations = read_table(infile_original_dir+'human_orthologues.tsv')
            ensg_idx = 4
    ## catelogue the results
    ensg_mito_ribo_dict=process_dict(results,3)
    # if args.species!="hsapiens":
    #     ensg_mito_ribo_dict = convert_dict_to_dict(all_mito_ribo_genes,)
    all_mito_ribo_genes = list(ensg_mito_ribo_dict.keys())
    print('\tfound',len(all_mito_ribo_genes),'mito or ribo genes')
    print('\t\t',all_mito_ribo_genes[:3])
    final_clust_gene_ids = []
    final_clust_gene_idxs = []
    for i in range(1,len(annotations)):
        temp_gene = annotations[i][ensg_idx]
        #temp_idx = clust_gene_idxs[i]
        if quick_search(ensg_mito_ribo_dict,temp_gene):
            print('\texcluding:',annotations[i][1])
            ## here we don't include the mito and ribo genes in the final list
        else:
            if int(annotations[i][0])-1 in usable_indices:
                #print("\tproblem mapping",temp_original_id,"but we'll include it")
                final_clust_gene_ids.append(ID_list[int(annotations[i][0])-1])
                final_clust_gene_idxs.append(int(annotations[i][0])-1)
    clust_gene_ids = list(set(final_clust_gene_ids))
    clust_gene_idxs = list(set(final_clust_gene_idxs))
    usable_indices = clust_gene_idxs
    ### 
    print(len(usable_indices),"currenlty usable indices")
    temp_usable_indices = list(set(usable_indices))
    return(temp_usable_indices, clust_gene_idxs, clust_gene_ids)


##########################################################
def remove_sparse(full_expression, temp_usable_indices):
## remove genes that are only expressed in a very small number of cells
    remove_sparse=True
    if remove_sparse and args.clust_on_genes == None:
        print('removing genes that are only observed in 5 percent or less of cells, or greater than 50 cells, whichever is lower:')
        cutoff = min([np.shape(full_expression)[1]*.05,50])
        print("\t",cutoff,"cells")
        usable_indices = []
        for i in temp_usable_indices:
            num_seen = np.sum(np.array(full_expression[i,:]>0,dtype=int))
            if num_seen >= cutoff:
                usable_indices.append(i)
        print(len(temp_usable_indices)-len(usable_indices),'genes removed because of sparse expression')
        print(len(usable_indices),'used for clustering')
        clust_gene_idxs = usable_indices[:]
        clust_gene_ids = [IDlist[idx] for idx in usable_indices]
    return(usable_indices, clust_gene_idxs, clust_gene_ids)
##########################################################
##########################################################
################# end feature selection ##################
##########################################################
##########################################################
##########################################################

## /SECTION 2

############################
def non_dispersion_feat_select(remove_ribo_mito = True):

    if args.neg_cor_clust != None:
        get_neg_cor_clust(neg_cor_clust)

    if remove_ribo_mito:
        usable_indices, clust_gene_idxs, clust_gene_ids = rm_ribo_mito(species, infile_original_dir)

    if not clust_on_genes:
        usable_indices, clust_gene_idxs, clust_gene_ids = remove_sparse(full_expression, usable_indices)
    return(usable_indices, clust_gene_idxs, clust_gene_ids)

############################

if __name__ == '__main__':

    non_dispersion_feat_select(neg_cor_clust = args.neg_cor_clust,
                               neg_cor_percentile_cutoff = args.percentile_cutoff,
                               out_dir = args.out_dir,
                               remove_ribo_mito = args.remove_ribo_mito)


#!/usr/bin/env python3
import argparse
import fileinput
import os
import sys
from copy import deepcopy
import numpy as np
import pickle
from sklearn import metrics
import networkx as nx

try:
    from pyminer.common_functions import *
    from pyminer.pyminer_ensembl_rest import get_genes_from_locus, EnsemblRestClient
except:
    from common_functions import *
    from pyminer_ensembl_rest import get_genes_from_locus, EnsemblRestClient
#####################################################################

# ## basic function library
# def read_file(tempFile,linesOraw='lines',quiet=False):
#     if not quiet:
#         print('reading',tempFile)
#     f=open(tempFile,'r')
#     if linesOraw=='lines':
#         lines=f.readlines()
#         for i in range(0,len(lines)):
#             lines[i]=lines[i].strip('\n')
#     elif linesOraw=='raw':
#         lines=f.read()
#     f.close()
#     return(lines)

# def make_file(contents,path):
#     f=open(path,'w')
#     if isinstance(contents,list):
#         f.writelines(contents)
#     elif isinstance(contents,str):
#         f.write(contents)
#     f.close()

    
# def flatten_2D_table(table,delim):
#     #print(type(table))
#     if str(type(table))=="<class 'numpy.ndarray'>":
#         out=[]
#         for i in range(0,len(table)):
#             out.append([])
#             for j in range(0,len(table[i])):
#                 try:
#                     str(table[i][j])
#                 except:
#                     print(table[i][j])
#                 else:
#                     out[i].append(str(table[i][j]))
#             out[i]=delim.join(out[i])+'\n'
#         return(out)
#     else:
#         for i in range(0,len(table)):
#             for j in range(0,len(table[i])):
#                 try:
#                     str(table[i][j])
#                 except:
#                     print(table[i][j])
#                 else:
#                     table[i][j]=str(table[i][j])
#             table[i]=delim.join(table[i])+'\n'
#     #print(table[0])
#         return(table)

# def strip_split(line, delim = '\t'):
#     return(line.strip('\n').split(delim))

# def make_table(lines,delim, num_type = float):
#     #print(num_type)
#     for i in range(0,len(lines)):
#         lines[i]=lines[i].strip()
#         lines[i]=lines[i].split(delim)
#         for j in range(0,len(lines[i])):
#             try:
#                 float(lines[i][j])
#             except:
#                 lines[i][j]=lines[i][j].replace('"','')
#             else:
#                 if num_type == float:
#                     lines[i][j]=float(lines[i][j])
#                 elif num_type == int:
#                     lines[i][j]=int(float(lines[i][j]))
#                 else:
#                     lines[i][j]=num_type(lines[i][j])
#     return(lines)


# def strip_split(line, delim = '\t'):
#     return(line.strip('\n').split(delim))

# def get_file_path(in_path):
#     in_path = in_path.split('/')
#     in_path = in_path[:-1]
#     in_path = '/'.join(in_path)
#     return(in_path+'/')

# def read_table(file, sep='\t',num_type=float):
#     return(make_table(read_file(file,'lines'),sep,num_type=num_type))
    
# def write_table(table, out_file, sep = '\t'):
#     make_file(flatten_2D_table(table,sep), out_file)

# def import_dict(f):
#     f=open(f,'rb')
#     d=pickle.load(f)
#     f.close()
#     return(d)

# def save_dict(d,path):
#     f=open(path,'wb')
#     pickle.dump(d,f)
#     f.close()

#####################################################################

#####################################################################

def get_contigs(in_file):
    temp_contigs = []
    first = True
    for line in fileinput.input(in_file):
        if first:
            first = False
        else:
            temp_line = strip_split(line)
            if temp_line[0] not in temp_contigs:
                temp_contigs.append(temp_line[0])
    fileinput.close()
    print("found these contigs:")
    for contig in temp_contigs:
        print("\t",contig)
    return(temp_contigs)


def populate_contig_dict(contig_dict, in_file):
    all_binding_sites = read_table(in_file)
    for i in range(1,len(all_binding_sites)):
        site = all_binding_sites[i]
        temp_cont = contig_dict[site[0]]
        temp_cont.append(site)
        contig_dict[site[0]] = temp_cont
    return(contig_dict)


def get_current_edges(cur_row, cur_index):
    cur_row = np.array(cur_row)
    indices = np.where(cur_row == True)[0]
    all_edges = []
    for index in indices:
        if index > cur_index:
            all_edges.append((cur_index,index))
    return(all_edges)


def get_insulator_sites_from_adj_mat(adj_mat):
    print(adj_mat)
    print("constructing graph")
    G = nx.Graph()
    G.add_nodes_from(list(range(adj_mat.shape[0])))
    for i in range(adj_mat.shape[0]):
        G.add_edges_from(get_current_edges(adj_mat[i,:],i))
    print("identifying connected components")
    comps = list(nx.connected_components(G))
    for i in range(len(comps)):
        comps[i] = sorted(list(map(int,list(comps[i]))))
    return(comps)


def get_rearranged_insulator_sites(non_gene_contig_list, insulator_site_lists):
    ## go through and replace the indices with the actual insulator site entry
    for i in range(len(insulator_site_lists)):
        for j in range(len(insulator_site_lists[i])):
            temp_idx = insulator_site_lists[i][j]
            insulator_site_lists[i][j] = non_gene_contig_list[temp_idx]
    return(insulator_site_lists)


def get_contig_gene_dict(lib_dir):
    ## read in the gene_contig_loci.tsv file
    ## "Gene stable ID","Gene name","Chromosome/scaffold name","Gene start (bp)","Gene end (bp)" <- has this header
    gene_annotations = np.array(read_table(os.path.expanduser(os.path.join(lib_dir, "gene_contig_loci.tsv")), num_convert = int))[1:,:]
    contigs = sorted(list(set(gene_annotations[:,2].tolist())))
    for i in range(len(contigs)):
        try:
            float(contigs[i])
        except:
            pass
        else:
            contigs[i] = str(int(float(contigs[i])))
    print("found these contigs in the gene reference file:\n\t",contigs)
    gene_contig_dict = {}
    for contig in contigs:
        ## find the indices for this contig
        #print("\n",contig)
        contig_indices = np.where(gene_annotations[:,2] == contig)[0]
        #print("contig_indices:",contig_indices)
        gene_contig_dict[contig] = gene_annotations[contig_indices,:]
        #print(gene_contig_dict[contig])
    return(gene_contig_dict)

##################################################
class gene_contig_lookup():
    def __init__(self,gene_cont_array):
        self.gene_cont_array = gene_cont_array
        print(self.gene_cont_array)
        
        self.processess_gene_cont_array()


    def get_return_lists(self):
        ## this is to match the chip binding site format
        ## chr, start, end, element(in this case gene), center, score(in this case "NA"), element_type(in this case "gene")
        self.return_list = []
        print("setting return lists")
        for i in range(self.gene_cont_array.shape[0]):
            temp_cont = self.gene_cont_array[i][2]
            start = self.start_bp_array[i]
            end = self.end_bp_array[i]
            center = self.center_bp_array[i]
            element = self.gene_cont_array[i][0]#+","+self.gene_cont_array[i][1]
            self.return_list.append([temp_cont,start,end,element,center,"NA","gene"])
            #print(self.return_list[-1])
        return


    def processess_gene_cont_array(self):
        temp_loci = np.array(self.gene_cont_array[:,np.array([3,4])],dtype = float)
        self.start_bp_array = np.min(temp_loci, axis = 1)
        self.end_bp_array = np.max(temp_loci, axis = 1)
        self.center_bp_array = np.mean(temp_loci, axis=1)
        self.gene_aray=self.gene_cont_array[:,0]
        self.symbol_array = self.gene_cont_array[:,2]
        self.contig_array = self.gene_cont_array[:,2]
        self.get_return_lists()
        return


    def check_for_uniform_direction(self,up1,up2,down1,down2):
        if up1 != up2:
            return(False)
        if up2 != down1:
            return(False)
        if down1 != down2:
            return(False)
        return(True)


    def get_genes(self, upstream_bp, downstream_bp):
        ## return all of the genes that have any degree of overlap with these start end end locations
        up_1_direction = np.array((self.start_bp_array-upstream_bp) >= 0, dtype = int)
        up_2_direction = np.array((self.end_bp_array-upstream_bp) >= 0, dtype = int)
        down_1_direction = np.array((self.start_bp_array-downstream_bp) >= 0, dtype = int)
        down_2_direction = np.array((self.end_bp_array-downstream_bp) >= 0, dtype = int)
        gene_loci = []
        sum_direction_array = up_1_direction+up_2_direction+down_1_direction+down_2_direction
        ## these are all of the indices where some elements are upstream, or downstream of the locus at hand, but they're not all uniformly up or down (0 or 4). This means that there is some locus overlap
        gene_indices = np.where((sum_direction_array>0) & (sum_direction_array<4))[0].tolist()
        for idx in gene_indices:
            gene_loci.append(self.return_list[idx])
        # for i in range(up_1_direction.shape[0]):
        #     ## if all of the elements are not uniformly positive or negative, it means that there is some
        #     ## overlap with this gene and the two loci provided
        #     if not self.check_for_uniform_direction(up_1_direction[i],up_2_direction[i],down_1_direction[i],down_2_direction[i]):
        #         # print(i)
        #         # print(self.return_list[i])
        #         gene_loci.append(self.return_list[i])
        return(gene_loci)

##################################################
def populate_contig_with_genes(all_insulator_sites, gene_contig_array):
    ## the insulator sites are a list of lists with all binding sites that should be linked together
    ## we need to figure out which genes to put in the chuncks in between those insulator sites, then add them to interveining lists of lists
    ## We'll iterate to the locations BETWEEN the insulator sites & populate a new array with genes that have
    if len(all_insulator_sites)==[]:
        if len(gene_contig_array)>0:
            return(gene_contig_array)
    try:
        all_insulator_sites[0][0][0]
    except:
        return()
        print(all_insulator_sites)
        print(gene_contig_array)
    # if all_insulator_sites[0][0][0]=="chrCHR_HG107_PATCH":
    #     print(all_insulator_sites)
    #     print(gene_contig_array)
    if type(gene_contig_array)==str:#"None":
        return(all_insulator_sites) 

    ###################
    ## prepare the gene contig for lookup 
    gene_lookup_obj = gene_contig_lookup(gene_contig_array)
    rearranged_contig = []

    #######################################################################
    ## first make a temprory pseudo-contig with only the insulator sites. We're mainly doing this to get the boundaries
    all_insulator_sites_contig = contig_obj("dummy",all_insulator_sites, finalize = False)

    ##### take care of the special case in which we're looking at the first locus - before the first insulator site
    ## first append the 
    temp_genes = gene_lookup_obj.get_genes(0,all_insulator_sites_contig.contig_tads[0].start)
    # print("\ngetting genes for beginning:")
    # print("\t",0,all_insulator_sites_contig.contig_tads[0].start)
    # for element in temp_genes:
    #     print(element)
    if len(temp_genes) > 0:
        rearranged_contig.append(temp_genes)
    ## append the first insulator locus
    rearranged_contig.append(all_insulator_sites[0])
    #####
    #sys.exit("done setting first set of genes")

    ## now take care of the middle
    for i in range(1,len(all_insulator_sites)-1):
        if i%100 == 0:
            print("\t",i,"/",len(all_insulator_sites))
        ## get the start and end location of this locus
        ################################################
        temp_start = all_insulator_sites_contig.contig_tads[i-1].end
        temp_end = all_insulator_sites_contig.contig_tads[i].start
        temp_genes = gene_lookup_obj.get_genes(temp_start, temp_end)
        rearranged_contig.append(temp_genes)
        ## then append the insulator locus
        rearranged_contig.append(all_insulator_sites[i])
        ################################################

    ## take care of the end - i.e. after the last insulator site
    rearranged_contig.append(all_insulator_sites[-1])
    temp_genes = gene_lookup_obj.get_genes(all_insulator_sites_contig.contig_tads[-1].end,9999999999)
    if len(temp_genes)>0:
        rearranged_contig.append(temp_genes)
    return(rearranged_contig)


def get_start_end_array(start_site, end_site):
    start_end_array = np.zeros((len(start_site),2))
    for i in range(0,len(start_site)):
        start_end_array[i,0] = start_site[i]
        start_end_array[i,1] = end_site[i]
    return(start_end_array)


def get_overlap_adj_mat(start_end_array):
    print("getting overlap_adj_mat")
    overlap_adj_mat = np.array(np.zeros((start_end_array.shape[0],start_end_array.shape[0])),dtype = bool)

    for i in range(start_end_array.shape[0]):
        ## return all of the genes that have any degree of overlap with these start end end locations
        upstream_bp = start_end_array[i,0]
        downstream_bp = start_end_array[i,1]
        up_1_direction = np.array((start_end_array[:,0]-upstream_bp) >= 0, dtype = int)## are the other genes' start downstream of the current gene's start
        up_2_direction = np.array((start_end_array[:,1]-upstream_bp) >= 0, dtype = int)## are the other genes' end downstream of the current gene's start
        down_1_direction = np.array((start_end_array[:,0]-downstream_bp) >= 0, dtype = int)## are the other genes' start downstream of the current gene's end
        down_2_direction = np.array((start_end_array[:,1]-downstream_bp) >= 0, dtype = int)## are the other genes' end downstream of the current gene's end
        gene_loci = []
        sum_direction_array = up_1_direction+up_2_direction+down_1_direction+down_2_direction
        ## these are all of the indices where some elements are upstream, or downstream of the locus at hand, but they're not all uniformly up or down (0 or 4). This means that there is some locus overlap
        gene_indices = np.where((sum_direction_array>0) & (sum_direction_array<4))[0]
        #print("\n")
        # for temp_array in [up_1_direction,up_2_direction,down_1_direction,down_2_direction]:
        #     print(temp_array)
        #print("sum:",sum_direction_array)
        #print(gene_indices)
        overlap_adj_mat[i,gene_indices]=True
    return(overlap_adj_mat)


def combine_adj_mats(adj_mat_centers, adj_mat_overlap):
    print("combining adj mats")
    print("centers\n",adj_mat_centers)
    print("overlap\n",adj_mat_overlap)
    if adj_mat_overlap.shape != adj_mat_centers.shape:
        print("adj_mat_centers.shape",adj_mat_centers.shape)
        print("adj_mat_overlap.shape",adj_mat_overlap.shape)

    final_adj_mat = adj_mat_centers + adj_mat_overlap
    if final_adj_mat.shape != adj_mat_centers.shape:
        print("final_adj_mat.shape",final_adj_mat.shape)
        print("adj_mat_centers.shape",adj_mat_centers.shape)
        print("adj_mat_overlap.shape",adj_mat_overlap.shape)
        sys.exit()
    return(final_adj_mat)

def get_center_adj_mat(center_site, dist_cutoff):
    center_site = np.array(center_site).reshape(-1, 1)
    dist_mat = np.abs(metrics.pairwise_distances(center_site, center_site))
    
    ## boolean if centers are within 150
    adj_mat_centers = dist_mat <= dist_cutoff
    return(adj_mat_centers)


def get_insulator_clusters_from_binding_sites(full_contig_list, lib_dir, contig_gene_array, dist_cutoff = 150):
    print("\nfinding all insulator clusters")
    non_gene_contig_list = []
    center_site = []
    start_site = []
    end_site = []

    ## filter the input for nly the non-gene elements
    for i in range(len(full_contig_list)):
        if full_contig_list[i][-1]!="gene":
            non_gene_contig_list.append(full_contig_list[i])
            center_site.append(float(full_contig_list[i][4]))
            start_end = [float(full_contig_list[i][1]),float(full_contig_list[i][2])]
            start_site.append(min(start_end))
            end_site.append(max(start_end))


    if len(non_gene_contig_list) > 1:
        ## if this contig actually has insulator elements
        print("getting adj_mat_centers")
        adj_mat_centers = get_center_adj_mat(center_site, dist_cutoff)
        
        ## now double check for overlap of any sites
        start_end_array = get_start_end_array(start_site, end_site)
        adj_mat_overlap = get_overlap_adj_mat(start_end_array)

        ## now combine the adjacency matrices
        adj_mat = combine_adj_mats(adj_mat_centers, adj_mat_overlap)
        print(adj_mat)

        insulator_site_indices = get_insulator_sites_from_adj_mat(adj_mat)
        all_insulator_sites = get_rearranged_insulator_sites(non_gene_contig_list, insulator_site_indices)
        for insulator_site in all_insulator_sites:
            print("\n") 
            for site in insulator_site:
                print("\t",site)
    else:
        all_insulator_sites = []
    rearranged_contig = populate_contig_with_genes(all_insulator_sites, contig_gene_array)
    return(rearranged_contig)


def get_temp_contig_gene_array(contig_gene_dict, contig):
    if contig in contig_gene_dict:
        print("found",contig)
        print(contig_gene_dict[contig])
        return(contig_gene_dict[contig])
    elif contig[3:] in contig_gene_dict:
        print("found",contig[3:])
        return(contig_gene_dict[contig[3:]])
    elif 'chr'+contig in contig_gene_dict:
        print("found",'chr'+contig)
        return(contig_gene_dict['chr'+contig])
    else:
        print("\n\ncouldn't find",contig,"in the gene reference file\n\n")
        return("None")



def filter_binding_site_dict(contig_dict, lib_dir, verbose = False, do_graph_clust = True):
    ## go through and figure out where the boundaries are
    all_contigs = sorted(list(contig_dict.keys()))
    rearranged_contig_dict = {}
    contig_gene_dict = get_contig_gene_dict(lib_dir)
    for contig in all_contigs:
        print("rearranging",contig)
        rearranged_contig = []
        
        ## figure out if we're starting with an insulator site or genes 
        temp_cont = contig_dict[contig]
        if do_graph_clust:
            rearranged_contig = get_insulator_clusters_from_binding_sites(temp_cont, lib_dir, get_temp_contig_gene_array(contig_gene_dict,contig))
        else:
            ##########################################################################
            if temp_cont[0][-1] == "gene":
                in_genes = True
            else:
                in_genes = False

            ## set up the loop
            cur_elements = []## this will hold the temporary elements of the insulator or gene set
            previous_locus = 0
            for i in range(len(temp_cont)):
                ## condition to find switch
                if in_genes:
                    if temp_cont[i][-1]=="gene":
                        ## in genes & continuing
                        cur_elements.append(temp_cont[i])
                    else:
                        if verbose:
                            print('\n')
                            for el in cur_elements:
                                print('\t',el)
                        ## in genes & not continuing
                        rearranged_contig.append(cur_elements)
                        ## re-set and 
                        in_genes = False
                        cur_elements = [temp_cont[i]]
                if not in_genes:
                    if temp_cont[i][-1]!="gene":
                        ## not in genes and continuing
                        cur_elements.append(temp_cont[i])
                    else:
                        if verbose:
                            print('\n')
                            for el in cur_elements:
                                print('\t',el)
                        ## not in genes and not continuing
                        rearranged_contig.append(cur_elements)
                        ## change_in_genes to True
                        in_genes = True
                        cur_elements = [temp_cont[i]]
        ##########################################################################
        rearranged_contig_dict[contig] = rearranged_contig

    return(rearranged_contig_dict)


class locus():
    def __init__(self,
                 locus_clump):
        self.contig=""
        self.start = None
        self.end = None
        self.genes = []
        self.annotations = ""
        self.locus_type = "between_insulators"
        ## for genes we'll set the gene list in the first pass, but not the boundaries
        ## for insulator sites, we'll set the boundaries, but not the genes in the first pass
        ## this variable will help us keep track of whether or not we're finished processing the given locus
        self.finished = False
        self.process_locus(locus_clump)


    def process_locus(self, locus_clump):
        if locus_clump == []:
            print("Nothing in this locus")
            return(None)
        # try:
        #     locus_clump[0][0]
        # except:
        #     print(locus_clump)
        self.contig = locus_clump[0][0]
        if locus_clump[0][-1] == "gene":
            self.locus_type = "gene"
        else:
            self.locus_type = "insulator"
        
        if self.locus_type == "gene":
            ## here we 
            self.process_gene(locus_clump)
        else:
            ##
            self.process_insulator(locus_clump)
        self.add_annotations(locus_clump)


    def add_annotations(self,locus_clump):
        annotation_list = []
        for i in range(len(locus_clump)):
            annotation_list.append(','.join(list(map(str,locus_clump[i]))))
        annotation_list = '|'.join(annotation_list)
        self.annotations = annotation_list
        return


    def process_gene(self, locus_clump):
        ## we won't set the boundaries, just the genes
        print("Found genes")
        for i in range(len(locus_clump)):
            self.genes.append(locus_clump[i][3])
        return


    def process_insulator(self, locus_clump):
        ## we won't set the genes, just the boundaries
        all_positions = []
        for i in range(0,len(locus_clump)):
            #print(locus_clump[i])
            all_positions.append(locus_clump[i][1])
            all_positions.append(locus_clump[i][2])
        self.start = min(all_positions)
        self.end = max(all_positions)
        #self.genes = sorted(list(set(self.genes + get_genes_from_locus(self.contig, self.start, self.end))))
        return


    def set_loci(self, upstream = None, downstream = None):
        ## check that upstream and downstream are insulators
        ## if not, print descriptive error
        if upstream == None:
            self.start = 0
        else:
            self.start = upstream.end+1
        if downstream == None:
            self.end = 99999999999
        else:
            self.end = downstream.start-1
        ## also set the contig if it's between insulator loci
        if self.locus_type == "between_insulators":
            if upstream is not None:
                self.contig = upstream.contig
            elif downstream is not None:
                self.contig = downstream.contig
            else:
                sys.exit("There's something seriously wrong with this contig")
            ## sanity check that the upstream ends upstream 
            if self.start > self.end:
                print("\n\n\nself.start < self.end:",self.start < self.end)
                print("self.start",self.start)
                print("self.end" , self.end)
                print("Upstream:\n",upstream)
                print("Downstream:\n",downstream)
                sys.exit("something went wrong with getting the up and downstream loci. Look above^^")
        self.finished = True
        return()


    def set_genes(self, upstream = None, downstream = None):
        if self.locus_type == "gene":
            return
        ## now we look for genes in the upstream and downstream if they exist & set our current genes to these
        if upstream != None:
            if upstream.locus_type == "gene":
                self.genes += upstream.genes
            else:
                print("'\twe got an upstream insulator site next to another insulator site")
                print("\t\t",upstream,"\n")
                print("\t\t",self,"\n")
                print("\t\t",downstream)
                sys.exit()
        else:
            self.start = 0
        if downstream != None:
            if downstream.locus_type == "gene":
                self.genes += downstream.genes
            else:
                print("'\twe got a downstream insulator site next to another insulator site")
                print("\t\t",upstream,"\n")
                print("\t\t",self,"\n")
                print("\t\t",downstream)
                ex
                sys.exit()
        else:
            self.end = 99999999999
        self.finished = True
        return


    def update_up_and_down(self, upstream = None, downstream = None, update = None):
        if update == "loci":
            self.set_loci(upstream = upstream, downstream = downstream)
        elif update == "genes":
            self.set_genes(upstream = upstream, downstream = downstream)
        return


    def get_output_line(self):
        ## contig, start, end, genes, annotatons
        out_line = [self.contig, str(int(self.start)), str(int(self.end)), self.locus_type, str(','.join(deepcopy(self.genes))), self.annotations] 
        return(out_line)

    def get_non_gene_elements(self):
        ## contact ensembl to get the non-gene elements overlapping with this span of genome
        client = EnsemblRestClient()
        temp_contig = self.contig
        if temp_contig[:3]=="chr":
            temp_contig = temp_contig[3:]
        self.regulatory_elements = client.get_regulatory_elements_from_region(temp_contig, str(int(self.start)), str(int(self.end)), feat = ["regulatory"])
        temp_anno_list = []
        for r in self.regulatory_elements["regulatory"]:
            temp_anno_list.append(r["description"]+","+str(r["id"])+","+str(r["seq_region_name"])+":"+str(r["start"])+"-"+str(r["end"]))
        self.annotations = "|".join(temp_anno_list)
        print(self.annotations,"\n")
        return()

    def __str__(self):
        return('\t'.join(self.get_output_line()))


###########################################################
class contig_obj():
    def __init__(self, name, list_of_locus_clumps, finalize = True):
        self.name = name
        self.all_gene_entries = []
        self.contig_tads = []
        self.process_contig(list_of_locus_clumps)
        if finalize:
            self.finalize_loci()
            self.prepare_for_lookup()

    def process_contig(self, list_of_locus_clumps):
        print("processing locus clumps for:",self.name)
        if len(list_of_locus_clumps) > 0:
            for element_set in list_of_locus_clumps:
                print(element_set)
                self.contig_tads.append(locus(element_set))
                for element in element_set:
                    if element[-1]=='gene':
                        self.all_gene_entries.append(element)
        ## now we log the start and end loci of all elements for easy lookup later
        self.gene_start_end = np.zeros((len(self.all_gene_entries),2))
        for i in range(len(self.all_gene_entries)):
            sites = [int(self.all_gene_entries[i][1]),int(self.all_gene_entries[i][2])]
            self.gene_start_end[i,0] = min(sites)
            self.gene_start_end[i,1] = max(sites)
        else:
            print('\tno genes or insulator sites in this contig; skipping')
        self.all_gene_entries = np.array(self.all_gene_entries)
        return


    def get_nearest_gene(self, start, end):
        if self.gene_start_end.shape[0]==0:
            return([])
        start_dist = np.abs(start - self.gene_start_end)
        print(start_dist)
        end_dist = np.abs(end - self.gene_start_end)
        print(end_dist)
        nearest_distance_for_all_elements = np.min(np.minimum.reduce([start_dist, end_dist]), axis = 1)
        print(nearest_distance_for_all_elements)
        nearest_gene_idx = np.where(nearest_distance_for_all_elements == np.min(nearest_distance_for_all_elements))[0]
        print(nearest_gene_idx)
        temp_genes = []
        for idx in range(nearest_gene_idx.shape[0]):
            #print(self.all_gene_entries[nearest_gene_idx[idx]])
            temp_genes.append(self.all_gene_entries[nearest_gene_idx[idx]][3])
        temp_genes = list(set(temp_genes))
        return(temp_genes)


    def update_non_gene_genes(self):
        ## TODO convert to counting to allow for boundary skipping
        ## first set the empty ranges
        for i in range(0,len(self.contig_tads)):
            if self.contig_tads[i].locus_type != "gene":
                temp_upstream = None
                temp_downstream = None
                ## first find the first available upstream "gene" locus
                upstream_idx = i
                upstream_gene_clump_count = 0
                found_up_genes = False
                while upstream_idx > 0 and not found_up_genes:
                    upstream_idx -= 1
                    if self.contig_tads[upstream_idx].locus_type == "gene":
                        upstream_gene_clump_count += 1
                        temp_upstream = self.contig_tads[upstream_idx]
                        found_up_genes = True

                ## next find the first available downstream "gene" locus
                downstream_idx = i
                downstream_gene_clump_count = 0
                found_down_genes = False
                while downstream_idx < len(self.contig_tads)-1 and not found_down_genes:
                    downstream_idx += 1
                    #print(downstream_idx,len(self.contig_tads))
                    if self.contig_tads[downstream_idx].locus_type == "gene":
                        downstream_gene_clump_count += 1
                        temp_downstream = self.contig_tads[downstream_idx]
                        found_down_genes = True


                self.contig_tads[i].update_up_and_down(upstream = temp_upstream, downstream = temp_downstream, update="genes")
        return()


    def update_non_insulator_loci(self):
        ## this finds the boundaries of all the 
        for i in range(0,len(self.contig_tads)):
            if self.contig_tads[i].locus_type != "insulator":
                if i == 0:
                    temp_upstream = None
                else:
                    temp_upstream = self.contig_tads[i-1]
                if i == len(self.contig_tads)-1:
                    temp_downstream = None
                else:
                    temp_downstream = self.contig_tads[i+1]
            
                self.contig_tads[i].update_up_and_down(upstream = temp_upstream, downstream = temp_downstream, update="loci")

    def update_between_insulator_elements(self):
        if True:
            return
        for i in range(0,len(self.contig_tads)):
            if self.contig_tads[i].locus_type == "between_insulators":
                self.contig_tads[i].get_non_gene_elements()
        return


    def finalize_loci(self):
        print("finalizing loci for",self.name)
        self.update_non_insulator_loci()
        self.update_non_gene_genes()
        self.update_between_insulator_elements()
        return

    def prepare_for_lookup(self):
        self.start_end_loci = np.zeros((len(self.contig_tads), 2))
        for i in range(len(self.contig_tads)):
            self.start_end_loci[i,0]=self.contig_tads[i].start
            self.start_end_loci[i,1]=self.contig_tads[i].end
        ## in the case where a tad contains no genes or insulator sites, give it a null set so that it plays nice with everything else
        if len(self.contig_tads)==0:
            self.start_end_loci = np.array(([[0,9999999999]]))
        return

    def get_single_tad_idx_site(self, bp):
        """
        Takes in a single base number & gets the tad that contains it
        """
        abs_dist_mat = np.abs(self.start_end_loci-bp)
        min_abs_dist = np.min(abs_dist_mat, axis = 1)
        nearest_tad_idx = np.where(min_abs_dist == np.min(min_abs_dist))[0][0]
        return(nearest_tad_idx)


    def get_tad_for_site(self, bp_start, bp_end = None):
        """
        Takes a base pair start and end number as input and finds the tad locus that contains the input base pair number
        We allow for start and end sites for things like multi-base deltions, cnvs, etc that are not single point mutants
        """
        tad_idx_start = self.get_single_tad_idx_site(bp_start)
        if bp_end != None:
            tad_idx_end = self.get_single_tad_idx_site(bp_end)
        else:
            tad_idx_end = tad_idx_start
        ## plus one because the last tad idx is inclusive, while fancy subsetting is not inclusive of the last idx
        return(self.contig_tads[tad_idx_start:tad_idx_end+1])


    def get_genes_from_bp_site(self, bp_start, bp_end = None):
        tad_set = self.get_tad_for_site(bp_start = bp_start, bp_end = bp_end)
        temp_genes = []
        tad_annotations = []
        tad_list = []
        for tad in tad_set:
            temp_genes += tad.genes
            tad_annotations.append([tad.contig, tad.start, tad.end, tad.locus_type, tad.annotations])
            tad_list.append(tad)
        temp_genes = sorted(list(set(temp_genes)))
        return(temp_genes, tad_annotations, tad_list)

###########################################################

def process_rearranged_contigs_to_tads(contig_dict):
    print("\n\nprocessing contigs into tads\n")
    all_contigs = sorted(list(contig_dict.keys()))
    tad_dict = {}
    ## go through each contig
    for contig in all_contigs:
        print("\t",contig, len(contig_dict[contig]))
        tad_dict[contig] = contig_obj(contig, contig_dict[contig])
    return(tad_dict)


def process_bed(in_dir):
    in_file = os.path.expanduser(os.path.join(in_dir, "chip_tad_boundaries_hg38.bed"))
    if not os.path.isfile(in_file):
        print("in_file:",in_file)
        sys.exit("we coudn't find the input file!:"+in_file)
    #### we'll do a few passes
    ## first to make all the contigs
    contig_dict = {contig:[] for contig in get_contigs(in_file)}
    ## and populate it with the binding sites
    contig_dict = populate_contig_dict(contig_dict, in_file)
    ## rearrange into chunks of elements that belong together
    rearranged_contig_dict = filter_binding_site_dict(contig_dict, in_dir)
    ## then we'll get the intervals of the tad boundaries
    tad_dict = process_rearranged_contigs_to_tads(rearranged_contig_dict)
    out_file = os.path.expanduser(os.path.join(in_dir,'tad_dict.pkl'))
    print("\n\nwriting the output file:",out_file)
    save_dict(tad_dict, out_file)
    print("Done")
    # ## As a test, we'll look up some loci
    # print("\n\n\nchr1:1,000,000\n")
    # for tad in tad_dict["chr1"].get_tad_for_site(1000000):
    #     print('\t',tad)
    # print("\n\n\nchr1:1,000,000:2,000,000\n")
    # for tad in tad_dict["chr1"].get_tad_for_site(1000000,2000000):
    #     print('\t',tad)

    # print("\n\ngenes:\n",tad_dict["chr1"].get_genes_from_bp_site(1000000))
    # print("\n\ngenes:\n",tad_dict["chr1"].get_genes_from_bp_site(1000000,2000000))

    ### write the output


#####################################################################

def main(args):
    process_bed(args.stringdb_dir)
    return

if __name__ == '__main__':
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-stringdb_dir",'-sdb',
        help='The directory containing the StringDB action files. Note that this HAS to be the sorted bed file of CTFC/cohesin binding sites.',
        type = str,
        default = '~/bin/pyminer/lib/')

    args = parser.parse_args()
    main(args)